@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -323,7 +323,13 @@ class ActionsWithProject(QObject):
                 self._specification_dicts.setdefault(item_type, []).append(spec_dict)
         dags = self._dags()
         settings = make_settings_dict_for_engine(self._app_settings)
+        # Force local execution in headless mode
+        if not settings.get("engineSettings/remoteExecutionEnabled", "false") == "false":
+            settings["engineSettings/remoteExecutionEnabled"] = "false"
         selected = {name for name_list in self._args.select for name in name_list} if self._args.select else None
+        deselected = {name for name_list in self._args.deselect for name in name_list} if self._args.deselect else None
+        executed_items = set()
+        skipped_items = set()
         for dag in dags:
             item_names_in_dag = set(dag.nodes)
             if not nx.is_directed_acyclic_graph(dag):
@@ -332,13 +338,15 @@ class ActionsWithProject(QObject):
             item_dicts_in_dag = {
                 name: item_dict for name, item_dict in self._item_dicts.items() if name in item_names_in_dag
             }
-            if selected is not None:
-                execution_permits = {item_name: item_name in selected for item_name in dag.nodes}
-                selected = selected - item_names_in_dag
-            else:
-                execution_permits = {item_name: True for item_name in item_names_in_dag}
+            execution_permits = {
+                item_name: (selected is None or item_name in selected)
+                and (deselected is None or item_name not in deselected)
+                for item_name in item_names_in_dag
+            }
+            skipped_items |= {name for name, selected in execution_permits.items() if not selected}
             if all(not permitted for permitted in execution_permits.values()):
                 continue
+            executed_items |= {name for name, selected in execution_permits.items() if selected}
             engine_data = {
                 "items": item_dicts_in_dag,
                 "specifications": self._specification_dicts,
@@ -349,8 +357,9 @@ class ActionsWithProject(QObject):
                 "settings": settings,
                 "project_dir": self._project_dir,
             }
-            engine_server_address = self._app_settings.value("appSettings/engineServerAddress", defaultValue="")
-            engine_manager = make_engine_manager(engine_server_address)
+            # exec_remotely is forced to False (see above)
+            exec_remotely = settings.get("engineSettings/remoteExecutionEnabled", "false") == "true"
+            engine_manager = make_engine_manager(exec_remotely)
             try:
                 engine_manager.run_engine(engine_data)
             except EngineInitFailed as error:
@@ -363,9 +372,15 @@ class ActionsWithProject(QObject):
                     if data == str(SpineEngineState.FAILED):
                         return Status.ERROR
                     break
-        if selected:
+        selected_invalid = selected - executed_items if selected is not None else None
+        deselected_invalid = deselected - skipped_items if deselected is not None else None
+        if selected_invalid:
             self._logger.msg_warning.emit(
-                f"The following selected items didn't exist in the project: {', '.join(selected)}"
+                f"The following selected items don't exist in the project: {', '.join(selected_invalid)}"
+            )
+        if deselected_invalid:
+            self._logger.msg_warning.emit(
+                f"The following deselected items don't exist in the project: {', '.join(deselected_invalid)}"
             )
         return Status.OK
 
