@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -16,7 +16,7 @@ General helper functions and classes.
 :date:   10.1.2018
 """
 
-from enum import Enum, unique
+from enum import Enum, unique, auto
 import itertools
 import os
 import glob
@@ -247,9 +247,7 @@ def pyside2_version_check():
     qt_version is the Qt version used to compile PySide2 as string. E.g. "5.14.2"
     qt_version_info is a tuple with each version component of Qt used to compile PySide2. E.g. (5, 14, 2)
     """
-    if not (qt_version_info[0] == 5 and qt_version_info[1] == 14) and not (
-        qt_version_info[0] == 5 and qt_version_info[1] == 15
-    ):
+    if not (qt_version_info[0] == 5 and qt_version_info[1] in (14, 15)):
         print(
             f"""Sorry for the inconvenience but,
 
@@ -881,6 +879,19 @@ def select_conda_executable(parent, line_edit):
     line_edit.setText(answer[0])
 
 
+def select_certificate_directory(parent, line_edit):
+    """Shows file browser and inserts selected certificate directory to given line edit.
+
+    Args:
+        parent (QWidget, optional): Parent of QFileDialog
+        line_edit (QLineEdit): Line edit where the selected dir path will be inserted
+    """
+    answer = QFileDialog.getExistingDirectory(parent, "Select certificates directory", home_dir())
+    if not answer:
+        return
+    line_edit.setText(answer)
+
+
 def file_is_valid(parent, file_path, msgbox_title, extra_check=None):
     """Checks that given path is not a directory and it's a file that actually exists.
     In addition, can be used to check if the file name in given file path starts with
@@ -1401,6 +1412,21 @@ def bisect_chunks(current_data, new_data, key=None):
 
 
 class FetchParent:
+    @unique
+    class Init(Enum):
+        UNINITIALIZED = auto()
+        IN_PROGRESS = auto()
+        FINISHED = auto()
+        FAILED = auto()
+
+    fetch_token = None
+    _is_fetched = False
+    _is_busy_fetching = False
+    query = None
+    query_key = None
+    query_iterator = None
+    query_initialized = Init.UNINITIALIZED
+
     @property
     def fetch_item_type(self):
         """Returns the type of item to fetch, e.g., "object_class".
@@ -1428,9 +1454,47 @@ class FetchParent:
     def fetch_status_change(self):
         """Called when fetch status changes."""
 
+    @property
+    def is_fetched(self):
+        return self._is_fetched
+
+    def set_fetched(self, fetched):
+        """Sets the fetched status.
+
+        Args:
+            fetched (bool): whether parent has been fetched completely
+        """
+        self._is_fetched = fetched
+
+    @property
+    def is_busy_fetching(self):
+        return self._is_busy_fetching
+
+    def set_busy_fetching(self, busy):
+        """Sets the busy status.
+
+        Args:
+            busy (bool): whether the parent is busy
+        """
+        self._is_busy_fetching = busy
+
+    def reset_fetching(self, fetch_token):
+        """Resets fetch parent as if nothing was ever fetched.
+
+        Args:
+            fetch_token (object): current fetch token
+        """
+        self.fetch_token = fetch_token
+        self._is_fetched = False
+        self.query = None
+        self.query_iterator = None
+        self.query_initialized = FetchParent.Init.UNINITIALIZED
+        self.fetch_status_change()
+
 
 class ItemTypeFetchParent(FetchParent):
     def __init__(self, fetch_item_type):
+        super().__init__()
         self._fetch_item_type = fetch_item_type
 
     @property
@@ -1570,3 +1634,20 @@ class HTMLTagFilter(HTMLParser):
     def handle_starttag(self, tag, attrs):
         if tag == "br":
             self._text += "\n"
+
+
+def same_path(path1, path2):
+    """Checks if two paths are equal.
+
+    This is a lightweight version of os.path.samefile(): it doesn't check if the paths
+    point to the same file system object but rather takes into account file system
+    case-sensitivity and such.
+
+    Args:
+        path1 (str): a path
+        path2 (str): a path
+
+    Returns:
+        bool: True if paths point to the same
+    """
+    return os.path.normcase(path1) == os.path.normcase(path2)
