@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -92,6 +92,14 @@ def _make_argument_parser():
     parser.add_argument(
         "-s", "--select", action="append", help="select project item ITEM for execution", nargs="*", metavar="ITEM"
     )
+    parser.add_argument(
+        "-d",
+        "--deselect",
+        action="append",
+        help="deselect project item ITEM for execution (takes precendence over --select)",
+        nargs="*",
+        metavar="ITEM",
+    )
     return parser
 
 
