@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -18,6 +18,7 @@ A model for maps, used by the parameter_value editors.
 from copy import deepcopy
 from numbers import Number
 from itertools import takewhile
+import numpy
 from PySide2.QtCore import QAbstractTableModel, QModelIndex, Qt
 from PySide2.QtGui import QColor, QFont
 from spinedb_api import (
@@ -72,7 +73,7 @@ class MapModel(QAbstractTableModel):
             parent (QObject): parent object
         """
         super().__init__(parent)
-        self._rows = convert_map_to_table(map_value, empty=empty)
+        self._rows = _numpy_string_to_python_strings(convert_map_to_table(map_value, empty=empty))
         self._index_names = _gather_index_names(map_value)
         self._BOLD = QFont()
         self._BOLD.setBold(True)
@@ -82,7 +83,7 @@ class MapModel(QAbstractTableModel):
         """Appends a new column to the right."""
         if not self._rows:
             return
-        first = len(self._rows[0])
+        first = len(self._rows[0]) + 1
         last = first
         self.beginInsertColumns(QModelIndex(), first, last)
         self._rows = list(map(lambda row: row + [empty], self._rows))
@@ -119,7 +120,7 @@ class MapModel(QAbstractTableModel):
     def columnCount(self, index=QModelIndex()):
         """Returns the number of columns in this model."""
         if not self._rows:
-            return 1
+            return len(self._index_names) + 2
         return len(self._rows[0]) + 1
 
     def convert_leaf_maps(self):
@@ -201,10 +202,9 @@ class MapModel(QAbstractTableModel):
         Return:
             bool: True if insertion was successful, False otherwise
         """
-        if not self._rows:
-            return False
         self.beginInsertColumns(parent, column, column + count - 1)
-        self._rows = [row[:column] + count * [empty] + row[column:] for row in self._rows]
+        if self._rows:
+            self._rows = [row[:column] + count * [empty] + row[column:] for row in self._rows]
         self._index_names = self._index_names[:column] + count * [Map.DEFAULT_INDEX_NAME] + self._index_names[column:]
         self.endInsertColumns()
         return True
@@ -229,7 +229,7 @@ class MapModel(QAbstractTableModel):
             else:
                 row_before = len(self._rows[0]) * [empty]
         else:
-            row_before = [empty, empty]
+            row_before = (len(self._index_names) + 1) * [empty]
         inserted = list()
         for _ in range(count):
             inserted.append([deepcopy(x) if x is not empty else x for x in row_before])
@@ -237,9 +237,24 @@ class MapModel(QAbstractTableModel):
         self.endInsertRows()
         return True
 
+    def is_leaf_value(self, index):
+        """Checks if given model index contains a leaf value.
+
+        Args:
+            index (QModelIndex): index to check
+
+        Returns:
+            bool: True if index points to leaf value, False otherwise
+        """
+        row = index.row()
+        column = index.column()
+        if self._is_in_expanse(row, column):
+            return False
+        return column > 0 and column + 1 == _data_length(self._rows[row])
+
     def _is_in_expanse(self, row, column):
         """
-        Returns True, if given row and column is in the right or bottom 'expanding' zone
+        Returns True, if given row and column is in the right or bottom 'expanding' zone.
 
         Args:
             row (int): row index
@@ -248,7 +263,7 @@ class MapModel(QAbstractTableModel):
         Returns:
             bool: True if the cell is in the expanse, False otherwise
         """
-        if row == len(self._rows):
+        if not self._rows or row == len(self._rows):
             return True
         return column == len(self._rows[0])
 
@@ -276,7 +291,7 @@ class MapModel(QAbstractTableModel):
         Returns:
             bool: True if row is the expanse row, False otherwise
         """
-        return row == len(self._rows)
+        return not self._rows or row == len(self._rows)
 
     def removeColumns(self, column, count, parent=QModelIndex()):
         """
@@ -322,7 +337,7 @@ class MapModel(QAbstractTableModel):
     def reset(self, map_value):
         """Resets the model to given map_value."""
         self.beginResetModel()
-        self._rows = convert_map_to_table(map_value, empty=empty)
+        self._rows = _numpy_string_to_python_strings(convert_map_to_table(map_value, empty=empty))
         self._index_names = _gather_index_names(map_value)
         self.endResetModel()
 
@@ -367,7 +382,9 @@ class MapModel(QAbstractTableModel):
         row_index = index.row()
         column_index = index.column()
         if row_index == len(self._rows):
-            self.insertRow(row_index)
+            if not value:
+                return False
+            self.insertRow(row_index + 1)
             row = self._rows[row_index]
             for i in range(column_index + 1, len(row)):
                 row[i] = empty
@@ -378,6 +395,8 @@ class MapModel(QAbstractTableModel):
         else:
             row = self._rows[row_index]
         if column_index == len(row):
+            if not value:
+                return False
             self.append_column()
             row = self._rows[row_index]
         if value is None or (isinstance(value, str) and value.lower() in ("null", "none")):
@@ -485,7 +504,12 @@ def _reconstruct_map(tree):
             value = _reconstruct_map(value)
         indexes.append(key)
         values.append(value)
-    return Map(indexes, values)
+    if len(indexes) > 1:
+        first_type = type(indexes[0])
+        if any(not isinstance(i, first_type) for i in indexes[1:]):
+            raise ParameterValueFormatError(f"Index type mismatch.")
+    map_ = Map(indexes, values)
+    return map_
 
 
 def _data_length(row):
@@ -532,3 +556,15 @@ def _apply_index_names(map_value, index_names):
     for value in map_value.values:
         if isinstance(value, Map):
             _apply_index_names(value, index_names[1:])
+
+
+def _numpy_string_to_python_strings(rows):
+    """Converts instances of numpy.str_ to regular Python strings.
+
+    Args:
+        rows (list of list): table rows
+
+    Returns:
+        list of list: converted rows
+    """
+    return [[str(x) if isinstance(x, numpy.str_) else x for x in row] for row in rows]
