@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -21,15 +21,17 @@ import os
 from pathlib import Path
 import json
 import random
-from PySide2.QtCore import Signal
+from PySide2.QtCore import Signal, QCoreApplication
 from PySide2.QtGui import QColor
 import networkx as nx
-from spine_engine.exception import EngineInitFailed
+from spine_engine.exception import EngineInitFailed, RemoteEngineInitFailed
 from spine_engine.utils.helpers import create_timestamp, gather_leaf_data
 from .project_item.logging_connection import LoggingConnection, LoggingJump
-from spine_engine.spine_engine import ExecutionDirection, validate_single_jump
-from spine_engine.utils.helpers import shorten
+from spine_engine.spine_engine import validate_single_jump
+from spine_engine.utils.helpers import ExecutionDirection, shorten, get_file_size
 from spine_engine.utils.serialization import deserialize_path, serialize_path
+from spine_engine.server.util.zip_handler import ZipHandler
+from .server.engine_client import EngineClient
 from .metaobject import MetaObject
 from .helpers import (
     create_dir,
@@ -40,6 +42,7 @@ from .helpers import (
     load_local_project_data,
     merge_dicts,
     load_specification_local_data,
+    busy_effect,
 )
 from .project_upgrader import ProjectUpgrader
 from .config import (
@@ -50,6 +53,7 @@ from .config import (
     PROJECT_LOCAL_DATA_FILENAME,
     FG_COLOR,
     SPECIFICATION_LOCAL_DATA_FILENAME,
+    PROJECT_ZIP_FILENAME,
 )
 from .project_commands import SetProjectNameAndDescriptionCommand
 from .spine_engine_worker import SpineEngineWorker
@@ -312,8 +316,11 @@ class SpineToolboxProject(MetaObject):
         self.restore_project_items(items_dict, item_factories, silent=True)
         self._logger.msg.emit("Restoring connections...")
         connection_dicts = project_info["project"]["connections"]
-        for connection in map(self.connection_from_dict, connection_dicts):
+        connections = list(map(self.connection_from_dict, connection_dicts))
+        for connection in connections:
             self.add_connection(connection, silent=True)
+        for connection in connections:
+            connection.link.update_icons()
         self._logger.msg.emit("Restoring jumps...")
         jump_dicts = project_info["project"].get("jumps", [])
         for jump in map(self.jump_from_dict, jump_dicts):
@@ -454,7 +461,7 @@ class SpineToolboxProject(MetaObject):
         if name != specification.name and self.is_specification_name_reserved(specification.name):
             self._logger.msg_error.emit(f"Specification name {specification.name} already in use.")
             return False
-        if save_to_disk and not self.save_specification_file(specification):
+        if save_to_disk and not self.save_specification_file(specification, name):
             return False
         id_ = self.specification_name_to_id(name)
         self._specifications[id_] = specification
@@ -477,13 +484,14 @@ class SpineToolboxProject(MetaObject):
         self.specification_replaced.emit(name, specification.name)
         return True
 
-    def save_specification_file(self, specification):
+    def save_specification_file(self, specification, previous_name=None):
         """Saves the given project item specification.
 
         Save path is determined by specification directory and specification's name.
 
         Args:
             specification (ProjectItemSpecification): specification to save
+            previous_name (str, optional): specification's earlier name if it has been renamed/replaced
 
         Returns:
             bool: True if operation was successful, False otherwise
@@ -494,20 +502,46 @@ class SpineToolboxProject(MetaObject):
                 return False
             specification.definition_file_path = candidate_path
         local_data = specification.save()
-        if local_data:
-            local_data_path = Path(self.config_dir, PROJECT_LOCAL_DATA_DIR_NAME, SPECIFICATION_LOCAL_DATA_FILENAME)
-            if local_data_path.exists():
-                with open(local_data_path) as local_data_file:
-                    specification_local_data = json.load(local_data_file)
-                specification_local_data.setdefault(specification.item_type, {})[specification.name] = local_data
-            else:
-                specification_local_data = {specification.item_type: {specification.name: local_data}}
-            local_data_path.parent.mkdir(parents=True, exist_ok=True)
-            with open(local_data_path, "w") as local_data_file:
-                self._dump(specification_local_data, local_data_file)
+        self._update_specification_local_data_store(specification, local_data, previous_name)
         self.specification_saved.emit(specification.name, specification.definition_file_path)
         return True
 
+    def _update_specification_local_data_store(self, specification, local_data, previous_name):
+        """Updates the file containing local data of project's specifications.
+
+        Args:
+            specification (ProjectItemSpecification): specification
+            local_data (dict): local data serialized into dict
+            previous_name (str, optional): specification's earlier name if it has been renamed/replaced
+        """
+        local_data_path = Path(self.config_dir, PROJECT_LOCAL_DATA_DIR_NAME, SPECIFICATION_LOCAL_DATA_FILENAME)
+        if local_data_path.exists():
+            with open(local_data_path) as local_data_file:
+                specification_local_data = json.load(local_data_file)
+            if local_data:
+                data_by_item_type = specification_local_data.setdefault(specification.item_type, {})
+                if previous_name is not None:
+                    data_by_item_type.pop(previous_name, None)
+                data_by_item_type[specification.name] = local_data
+            else:
+                local_item_data = specification_local_data.get(specification.item_type)
+                if local_item_data is not None:
+                    name = specification.name if previous_name is None else previous_name
+                    previous_data = local_item_data.pop(name, None)
+                    if previous_data is None:
+                        return
+                    if not local_item_data:
+                        del specification_local_data[specification.item_type]
+                else:
+                    return
+        elif not local_data:
+            return
+        else:
+            specification_local_data = {specification.item_type: {specification.name: local_data}}
+        local_data_path.parent.mkdir(parents=True, exist_ok=True)
+        with open(local_data_path, "w") as local_data_file:
+            self._dump(specification_local_data, local_data_file)
+
     def _default_specification_file_path(self, specification):
         """Determines a path inside project directory to save a specification.
 
@@ -743,6 +777,7 @@ class SpineToolboxProject(MetaObject):
         connection.source_position = source_position
         connection.destination_position = destination_position
         self.connection_updated.emit(connection)
+        connection.link.update_icons()
 
     def jumps_for_item(self, item_name):
         """Returns jumps that have given item as source or destination.
@@ -952,12 +987,16 @@ class SpineToolboxProject(MetaObject):
         if self._engine_workers:
             self._logger.msg_error.emit("Execution already in progress.")
             return
+        job_id = self.prepare_remote_execution()
+        if not job_id:
+            self.project_execution_finished.emit()
+            return
         settings = make_settings_dict_for_engine(self._settings)
         darker_fg_color = QColor(FG_COLOR).darker().name()
         darker = lambda x: f'<span style="color: {darker_fg_color}">{x}</span>'
         for k, (dag, execution_permits) in enumerate(zip(dags, execution_permits_list)):
             dag_identifier = f"{k + 1}/{len(dags)}"
-            worker = self.create_engine_worker(dag, execution_permits, dag_identifier, settings)
+            worker = self.create_engine_worker(dag, execution_permits, dag_identifier, settings, job_id)
             if worker is None:
                 continue
             self._logger.msg.emit("<b>Starting DAG {0}</b>".format(dag_identifier))
@@ -973,7 +1012,7 @@ class SpineToolboxProject(MetaObject):
         for worker in self._engine_workers:
             worker.start()
 
-    def create_engine_worker(self, dag, execution_permits, dag_identifier, settings):
+    def create_engine_worker(self, dag, execution_permits, dag_identifier, settings, job_id):
         """Creates and returns a SpineEngineWorker to execute given *validated* dag.
 
         Args:
@@ -981,25 +1020,31 @@ class SpineToolboxProject(MetaObject):
             execution_permits (dict): mapping item names to a boolean indicating whether to execute it or skip it
             dag_identifier (str): A string identifying the dag, for logging
             settings (dict): project and app settings to send to the spine engine.
+            job_id (str): Job Id for remote execution
 
         Returns:
             SpineEngineWorker
         """
-        item_dicts = {}
-        specification_dicts = {}
         items = {name: item for name, item in self._project_items.items() if name in dag.nodes}
-        for name, project_item in items.items():
-            item_dicts[name] = project_item.item_dict()
-            spec = project_item.specification()
-            if spec is not None:
-                spec_dict = spec.to_dict().copy()
-                spec_dict["definition_file_path"] = spec.definition_file_path
-                specification_dicts.setdefault(project_item.item_type(), list()).append(spec_dict)
+        item_dicts = {name: project_item.item_dict() for name, project_item in items.items()}
         connections = {c.name: c for c in self._connections if {c.source, c.destination}.intersection(items)}
         connection_dicts = [c.to_dict() for c in connections.values()]
         jumps = {c.name: c for c in self._jumps if execution_permits.get(c.source, False)}
         jump_dicts = [c.to_dict() for c in jumps.values()]
         connections.update(jumps)
+        specs_by_type = {}
+        for project_item in items.values():
+            spec = project_item.specification()
+            if spec is not None:
+                specs_by_type.setdefault(project_item.item_type(), list()).append(spec)
+        for jump in jumps.values():
+            if jump.condition["type"] == "tool-specification":
+                spec = self.get_specification(jump.condition["specification"])
+                specs_by_type.setdefault("Tool", list()).append(spec)
+        specification_dicts = {
+            type_: [{**spec.to_dict(), "definition_file_path": spec.definition_file_path} for spec in specs]
+            for type_, specs in specs_by_type.items()
+        }
         data = {
             "items": item_dicts,
             "specifications": specification_dicts,
@@ -1008,10 +1053,9 @@ class SpineToolboxProject(MetaObject):
             "execution_permits": execution_permits,
             "items_module_name": "spine_items",
             "settings": settings,
-            "project_dir": self.project_dir,
+            "project_dir": self.project_dir.replace(os.sep, "/"),
         }
-        server_address = self._settings.value("appSettings/engineServerAddress", defaultValue="")
-        worker = SpineEngineWorker(server_address, data, dag, dag_identifier, items, connections, self._logger)
+        worker = SpineEngineWorker(data, dag, dag_identifier, items, connections, self._logger, job_id)
         return worker
 
     def _handle_engine_worker_finished(self, worker):
@@ -1021,6 +1065,7 @@ class SpineToolboxProject(MetaObject):
             "COMPLETED": [self._logger.msg_success, "completed successfully"],
         }
         outcome = finished_outcomes.get(worker.engine_final_state())
+        # print("project._handle_engine_worker_finished() worker state: %s"%outcome)
         if outcome is not None:
             outcome[0].emit(f"<b>DAG {worker.dag_identifier} {outcome[1]}</b>")
         if any(worker.engine_final_state() not in finished_outcomes for worker in self._engine_workers):
@@ -1035,6 +1080,10 @@ class SpineToolboxProject(MetaObject):
                 item.handle_execution_successful(direction, state)
             finished_worker.clean_up()
         self._engine_workers.clear()
+        # We could remove the transmitted project zip-file here if we want
+        # ZipHandler.remove_file(
+        #     os.path.abspath(os.path.join(self._project_dir, os.pardir, PROJECT_ZIP_FILENAME + ".zip"))
+        # )
         self.project_execution_finished.emit()
 
     def execute_selected(self, names):
@@ -1107,7 +1156,7 @@ class SpineToolboxProject(MetaObject):
             item (ProjectItem): item whose resources have changed
         """
         item_name = item.name
-        predecessor_names = {c.source for c in self._incoming_connections(item_name)}
+        predecessor_names = {c.source for c in self.incoming_connections(item_name)}
         successor_connections = self._outgoing_connections
         update_resources = self._update_predecessor
         trigger_resources = item.resources_for_direct_predecessors()
@@ -1128,7 +1177,7 @@ class SpineToolboxProject(MetaObject):
         """
         item_name = item.name
         successor_names = {c.destination for c in self._outgoing_connections(item_name)}
-        predecessor_connections = self._incoming_connections
+        predecessor_connections = self.incoming_connections
         update_resources = self._update_successor
         trigger_resources = item.resources_for_direct_successors()
         self._notify_resource_changes(
@@ -1201,7 +1250,7 @@ class SpineToolboxProject(MetaObject):
         """
         target_name = target_item.name
         if direction == ExecutionDirection.FORWARD:
-            connections = self._incoming_connections(target_name)
+            connections = self.incoming_connections(target_name)
             self._update_successor(target_item, connections, resource_cache={})
         else:
             connections = self._outgoing_connections(target_name)
@@ -1216,7 +1265,7 @@ class SpineToolboxProject(MetaObject):
         Returns:
             set of str: direct predecessor names
         """
-        return {c.source for c in self._incoming_connections(name)}
+        return {c.source for c in self.incoming_connections(name)}
 
     def successor_names(self, name):
         """Collects direct successor item names.
@@ -1262,7 +1311,7 @@ class SpineToolboxProject(MetaObject):
         """
         return self._outgoing_connections(name) + self._outgoing_jumps(name)
 
-    def _incoming_connections(self, name):
+    def incoming_connections(self, name):
         """Collects incoming connections.
 
         Args:
@@ -1293,7 +1342,7 @@ class SpineToolboxProject(MetaObject):
         Returns:
             set of Connection/Jump: incoming connections
         """
-        return self._incoming_connections(name) + self._incoming_jumps(name)
+        return self.incoming_connections(name) + self._incoming_jumps(name)
 
     def _update_successor(self, successor, incoming_connections, resource_cache):
         combined_resources = list()
@@ -1341,6 +1390,60 @@ class SpineToolboxProject(MetaObject):
     def settings(self):
         return self._settings
 
+    @busy_effect
+    def prepare_remote_execution(self):
+        """Pings the server and sends the project as a zip-file to server.
+
+        Returns:
+            str: Job Id if server is ready for remote execution, empty string if something went wrong or "1" if
+            local execution is enabled.
+        """
+        if not self._settings.value("engineSettings/remoteExecutionEnabled", defaultValue="false") == "true":
+            return "1"  # Something that isn't False
+        host, port, sec_model, sec_folder = self._toolbox.engine_server_settings()
+        if not host:
+            self._logger.msg_error.emit(
+                "Spine Engine Server <b>host address</b> missing. "
+                "Please enter host in <b>File->Settings->Engine</b>."
+            )
+            return ""
+        elif not port:
+            self._logger.msg_error.emit(
+                "Spine Engine Server <b>port</b> missing. " "Please select port in <b>File->Settings->Engine</b>."
+            )
+            return ""
+        self._logger.msg.emit(f"Connecting to Spine Engine Server at <b>{host}:{port}</b>")
+        try:
+            engine_client = EngineClient(host, port, sec_model, sec_folder)
+        except RemoteEngineInitFailed as e:
+            self._logger.msg_error.emit(
+                f"Server is not responding. {e}. " f"Check settings in <b>File->Settings->Engine</b>."
+            )
+            return ""
+        engine_client.set_start_time()  # Set start_time for upload operation
+        # Archive the project into a zip-file
+        dest_dir = os.path.join(self.project_dir, os.pardir)  # Parent dir of project_dir TODO: Find a better dst
+        self._logger.msg.emit(f"Squeezing project <b>{self.name}</b> into {PROJECT_ZIP_FILENAME}.zip")
+        QCoreApplication.processEvents()
+        try:
+            ZipHandler.package(src_folder=self.project_dir, dst_folder=dest_dir, fname=PROJECT_ZIP_FILENAME)
+        except Exception as e:
+            self._logger.msg_error.emit(f"{e}")
+            return ""
+        project_zip_file = os.path.abspath(os.path.join(self.project_dir, os.pardir, PROJECT_ZIP_FILENAME + ".zip"))
+        if not os.path.isfile(project_zip_file):
+            self._logger.msg_error.emit(f"Project zip-file {project_zip_file} does not exist")
+            return ""
+        file_size = get_file_size(os.path.getsize(project_zip_file))
+        self._logger.msg_warning.emit(f"Uploading project [{file_size}] ...")
+        QCoreApplication.processEvents()
+        _, project_dir_name = os.path.split(self.project_dir)
+        job_id = engine_client.upload_project(project_dir_name, project_zip_file)
+        t = engine_client.get_elapsed_time()
+        self._logger.msg.emit(f"Upload time: {t}. Job ID: <b>{job_id}</b>")
+        engine_client.close()
+        return job_id
+
     def tear_down(self):
         """Cleans up project."""
         if self._execution_in_progress:
