@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -16,18 +16,10 @@ QUndoCommand subclasses for modifying the project.
 :date:   12.2.2020
 """
 
-from enum import IntEnum, unique
 from PySide2.QtWidgets import QUndoCommand
 from spine_engine.project_item.connection import Connection, Jump
 
 
-@unique
-class Id(IntEnum):
-    """Id numbers for project commands."""
-
-    JUMP_CONDITION = 1
-
-
 class SpineToolboxCommand(QUndoCommand):
     def __init__(self):
         super().__init__()
@@ -410,15 +402,6 @@ class SetJumpConditionCommand(SpineToolboxCommand):
         self._previous_condition = jump.condition
         self.setText("change loop condition")
 
-    def id(self):
-        return Id.JUMP_CONDITION
-
-    def mergeWith(self, other):
-        if not isinstance(other, SetJumpConditionCommand) or self._jump is not other._jump:
-            return False
-        self._condition = other._condition
-        return True
-
     def redo(self):
         self._jump_properties.set_condition(self._jump, self._condition)
 
