@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -594,7 +594,8 @@ class ObjectItem(EntityItem):
             return {(db_map, id_) for db_map, rel_cls in db_map_rel_clss for id_ in rel_cls["relationship_ids"]}
         return {
             (db_map, id_)
-            for db_map, rel_cls in self._db_map_relationship_class_lists.values()
+            for class_list in self._db_map_relationship_class_lists.values()
+            for db_map, rel_cls in class_list
             for id_ in rel_cls["relationship_ids"]
         }
 
