@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -266,11 +266,19 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
         """Sets the filter invalid."""
         self._filter_timer.start()
 
+    def stop_invalidating_filter(self):
+        """Stops invalidating the filter."""
+        self._filter_timer.stop()
+
     def set_filter_class_ids(self, class_ids):
         if class_ids != self._filter_class_ids:
             self._filter_class_ids = class_ids
             self._invalidate_filter()
 
+    def clear_auto_filter(self):
+        self._auto_filter = {}
+        self._invalidate_filter()
+
     @Slot(str, dict)
     def set_auto_filter(self, field, values):
         """Updates and applies the auto filter.
@@ -316,8 +324,8 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
         Args:
             model (SingleParameterModel, EmptyParameterModel)
 
-        Returns:
-            list: tuples (model, row number) for each accepted row
+        Yields:
+            tuple: (model, row number) for each accepted row
         """
         if not self.filter_accepts_model(model):
             return ()
@@ -347,7 +355,8 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
             ids = {x["id"] for x in data}
             for model in self._models_with_db_map(db_map):
                 if model.entity_class_id in ids:
-                    self.sub_models.remove(model)
+                    i = self.sub_models.index(model)
+                    self.sub_models.pop(i).deleteLater()
         self._do_refresh()
         self.layoutChanged.emit()
 
@@ -377,15 +386,21 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
             db_map_data (dict): list of added dict-items keyed by DiffDatabaseMapping
         """
         for db_map, items in db_map_data.items():
+            db_map_single_models = [m for m in self.single_models if m.db_map is db_map]
+            existing_ids = set().union(*(m.item_ids() for m in db_map_single_models))
             items_per_class = self._items_per_class(items)
             for entity_class_id, class_items in items_per_class.items():
                 ids_committed = list()
                 ids_uncommitted = list()
                 for item in class_items:
-                    if item.get("commit_id") is not None:
-                        ids_committed.append(item["id"])
+                    is_committed = item.get("commit_id") is not None
+                    item_id = item["id"]
+                    if item_id in existing_ids:
+                        continue
+                    if is_committed:
+                        ids_committed.append(item_id)
                     else:
-                        ids_uncommitted.append(item["id"])
+                        ids_uncommitted.append(item_id)
                 self._add_parameter_data(db_map, entity_class_id, ids_committed, committed=True)
                 self._add_parameter_data(db_map, entity_class_id, ids_uncommitted, committed=False)
                 self._do_add_data_to_filter_menus(db_map, class_items)
@@ -404,6 +419,14 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
         return model
 
     def _add_parameter_data(self, db_map, entity_class_id, ids, committed):
+        """Creates new single model and resets it with the given parameter ids.
+
+        Args:
+            db_map (DiffDatabaseMapping): database map
+            entity_class_id (int): parameter's entity class id
+            ids (list of int): parameter ids
+            committed (bool): True if the ids have been committed, False otherwise
+        """
         if not ids:
             return
         if committed:
@@ -472,6 +495,8 @@ class CompoundParameterModel(FetchParent, CompoundWithEmptyTableModel):
     def db_map_id(self, index):
         sub_index = self.map_to_sub(index)
         sub_model = sub_index.model()
+        if sub_model is None:
+            return None, None
         return sub_model.db_map, sub_model.item_id(sub_index.row())
 
     def index_name(self, index):
