@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -68,8 +68,18 @@ class PivotModel:
         self._data.update(data)
 
     def add_to_model(self, data):
-        data.update(self._data)
-        self._data = data
+        """Adds data to model.
+
+        Args:
+            data (dict): pivot model data
+
+        Returns:
+            tuple: added row count and added column count
+        """
+        addable_data = {k: v for k, v in data.items() if v is not None or k not in self._data}
+        if not addable_data:
+            return 0, 0
+        self._data.update(addable_data)
         if not any(self.frozen_value):
             first = next(iter(self._data), None)
             if first is None:
@@ -186,6 +196,7 @@ class PivotModel:
         Args:
             row_mask (list)
             column_mask (list)
+
         Returns:
             list(list)
         """
