@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -17,13 +17,14 @@ Custom item delegates.
 """
 
 from numbers import Number
-from PySide2.QtCore import QModelIndex, QPoint, Qt, Signal, QSortFilterProxyModel, QRegExp
+from PySide2.QtCore import QModelIndex, QPoint, Qt, Signal
 from PySide2.QtWidgets import QStyledItemDelegate, QComboBox
 from PySide2.QtGui import QFontMetrics
 from spinedb_api import to_database
 from spinedb_api.parameter_value import join_value_and_type
+from ..mvcmodels.single_parameter_models import SingleParameterModel
 from ...widgets.custom_editors import CustomLineEditor, SearchBarEditor, CheckListEditor, ParameterValueLineEditor
-from ...mvcmodels.shared import PARSED_ROLE
+from ...mvcmodels.shared import PARSED_ROLE, DB_MAP_ROLE
 from ...widgets.custom_delegates import CheckBoxDelegate, RankDelegate
 from ...helpers import object_icon
 from ..mvcmodels.metadata_table_model_base import Column as MetadataColumn
@@ -250,9 +251,8 @@ class ParameterDelegate(QStyledItemDelegate):
         """Returns the db_map for the database at given index or None if not set yet."""
         model = index.model()
         header = model.horizontal_header_labels()
-        database = index.sibling(index.row(), header.index("database")).data()
-        db_map = next(iter(x for x in self.db_mngr.db_maps if x.codename == database), None)
-        if not db_map:
+        db_map = index.sibling(index.row(), header.index("database")).data(DB_MAP_ROLE)
+        if db_map is None:
             self.parent().msg_error.emit("Please select database first.")
         return db_map
 
