@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -108,7 +108,8 @@ class EntityQGraphicsView(CustomQGraphicsView):
         self.selected_items = selected_objs + selected_rels
         self.graph_selection_changed.emit({"object": selected_objs, "relationship": selected_rels})
         default_data = selected_items[0].default_parameter_data() if len(selected_items) == 1 else {}
-        self._spine_db_editor.set_default_parameter_data(default_data)
+        default_db_map = selected_items[0].first_db_map if len(selected_items) == 1 else None
+        self._spine_db_editor.set_default_parameter_data(default_data, default_db_map)
 
     def connect_spine_db_editor(self, spine_db_editor):
         self._spine_db_editor = spine_db_editor
