@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -15,9 +15,10 @@ Custom QTableView classes that support copy-paste and the like.
 :author: M. Marin (KTH)
 :date:   18.5.2018
 """
+from dataclasses import replace
 
 from PySide2.QtCore import Qt, Signal, Slot, QTimer, QModelIndex, QPoint
-from PySide2.QtWidgets import QAction, QTableView, QMenu
+from PySide2.QtWidgets import QAction, QTableView, QHeaderView, QMenu
 from PySide2.QtGui import QKeySequence
 
 from .scenario_generator import ScenarioGenerator
@@ -29,11 +30,16 @@ from ..mvcmodels.pivot_table_models import (
 )
 from ..mvcmodels.metadata_table_model_base import Column as MetadataColumn
 from ...widgets.report_plotting_failure import report_plotting_failure
-from ...widgets.plot_widget import PlotWidget, _prepare_plot_in_window_menu
+from ...widgets.plot_widget import PlotWidget, prepare_plot_in_window_menu
 from ...widgets.custom_qtableview import CopyPasteTableView, AutoFilterCopyPasteTableView
 from ...widgets.custom_qwidgets import TitleWidgetAction
-from ...plotting import plot_selection, PlottingError, ParameterTablePlottingHints, PivotTablePlottingHints
-from ...helpers import preferred_row_height, rows_to_row_count_tuples
+from ...plotting import (
+    PlottingError,
+    ParameterTableHeaderSection,
+    plot_parameter_table_selection,
+    plot_pivot_table_selection,
+)
+from ...helpers import preferred_row_height, rows_to_row_count_tuples, DB_ITEM_SEPARATOR
 from .pivot_table_header_view import (
     PivotTableHeaderView,
     ParameterValuePivotHeaderView,
@@ -62,6 +68,9 @@ def _set_parameter_data(index, new_value):
 
 
 class ParameterTableView(AutoFilterCopyPasteTableView):
+    value_column_header: str = NotImplemented
+    """Either "default value" or "value". Used to identify the value column for advanced editing and plotting."""
+
     def __init__(self, parent):
         """
         Args:
@@ -75,11 +84,6 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         self._plot_separator = None
         self.pinned_values = []
 
-    @property
-    def value_column_header(self):
-        """Either "default value" or "value". Used to identify the value column for advanced editing and plotting."""
-        raise NotImplementedError()
-
     def connect_spine_db_editor(self, spine_db_editor):
         """Connects a Spine db editor to work with this view.
 
@@ -87,8 +91,10 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
              spine_db_editor (SpineDBEditor)
         """
         self._spine_db_editor = spine_db_editor
+        self.set_external_copy_and_paste_actions(spine_db_editor.ui.actionCopy, spine_db_editor.ui.actionPaste)
         self.populate_context_menu()
         self.create_delegates()
+        self.selectionModel().selectionChanged.connect(self._refresh_copy_paste_actions)
 
     def _make_delegate(self, column_name, delegate_class):
         """Creates a delegate for the given column and returns it.
@@ -120,14 +126,25 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         """Plots current index."""
         selection = self.selectedIndexes()
         try:
-            hints = ParameterTablePlottingHints()
-            plot_widget = plot_selection(self.model(), selection, hints)
+            plot_widget = self._plot_selection(selection)
         except PlottingError as error:
             report_plotting_failure(error, self._spine_db_editor)
         else:
             plot_widget.use_as_window(self.window(), self.value_column_header)
             plot_widget.show()
 
+    def _plot_selection(self, selection, plot_widget=None):
+        """Adds selected indexes to existing plot or creates a new plot window.
+
+        Args:
+            selection (Iterable of QModelIndex): a list of QModelIndex objects for plotting
+            plot_widget (PlotWidget, optional): an existing plot widget to draw into or None to create a new widget
+
+        Returns:
+            PlotWidget: a PlotWidget object
+        """
+        raise NotImplementedError()
+
     @Slot(QAction)
     def plot_in_window(self, action):
         """Plots current index in the window given by action's name."""
@@ -135,8 +152,7 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         plot_window = PlotWidget.plot_windows.get(plot_window_name)
         selection = self.selectedIndexes()
         try:
-            hints = ParameterTablePlottingHints()
-            plot_selection(self.model(), selection, hints, plot_window)
+            self._plot_selection(selection, plot_window)
             plot_window.raise_()
         except PlottingError as error:
             report_plotting_failure(error, self._spine_db_editor)
@@ -155,7 +171,8 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         self._menu.addAction("Filter by", self.filter_by_selection)
         self._menu.addAction("Filter excluding", self.filter_excluding_selection)
         self._menu.addSeparator()
-        self._menu.aboutToShow.connect(self._spine_db_editor.refresh_copy_paste_actions)
+        self._menu.addAction("Clear all filters", self._spine_db_editor.clear_all_filters)
+        self._menu.addSeparator()
         # Shortcuts
         remove_rows_action.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_Delete))
         remove_rows_action.setShortcutContext(Qt.WidgetShortcut)
@@ -177,7 +194,7 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         if is_value:
             plot_in_window_menu = QMenu("Plot in window")
             plot_in_window_menu.triggered.connect(self.plot_in_window)
-            _prepare_plot_in_window_menu(plot_in_window_menu)
+            prepare_plot_in_window_menu(plot_in_window_menu)
             self._menu.insertMenu(self._plot_separator, plot_in_window_menu)
         self._menu.exec_(event.globalPos())
         if is_value:
@@ -228,13 +245,14 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         # Get parameter data grouped by db_map
         db_map_typed_data = dict()
         model = self.model()
+        empty_model = model.empty_model
         for row in sorted(rows, reverse=True):
-            db_map = model.sub_model_at_row(row).db_map
-            if db_map is None:
-                # It's an empty model, just remove the row
-                _, sub_row = model._row_map[row]
-                model.empty_model.removeRow(sub_row)
+            sub_model = model.sub_model_at_row(row)
+            if sub_model is empty_model:
+                sub_row = model.sub_model_row(row)
+                sub_model.removeRow(sub_row)
                 continue
+            db_map = sub_model.db_map
             id_ = model.item_at_row(row)
             db_map_typed_data.setdefault(db_map, {}).setdefault(model.item_type, []).append(id_)
         model.db_mngr.remove_items(db_map_typed_data)
@@ -244,6 +262,11 @@ class ParameterTableView(AutoFilterCopyPasteTableView):
         super().rowsInserted(parent, start, end)
         self.resizeColumnsToContents()
 
+    @Slot(QModelIndex, QModelIndex)
+    def _refresh_copy_paste_actions(self, _, __):
+        """Enables or disables copy and paste actions."""
+        self._spine_db_editor.refresh_copy_paste_actions()
+
 
 class ObjectParameterTableMixin:
     def create_delegates(self):
@@ -258,9 +281,7 @@ class RelationshipParameterTableMixin:
 
 
 class ParameterDefinitionTableView(ParameterTableView):
-    @property
-    def value_column_header(self):
-        return "default_value"
+    value_column_header = "default_value"
 
     def create_delegates(self):
         super().create_delegates()
@@ -268,11 +289,13 @@ class ParameterDefinitionTableView(ParameterTableView):
         delegate = self._make_delegate("default_value", ParameterDefaultValueDelegate)
         delegate.parameter_value_editor_requested.connect(self._spine_db_editor.show_parameter_value_editor)
 
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class"""
+        raise NotImplementedError()
+
 
 class ParameterValueTableView(ParameterTableView):
-    @property
-    def value_column_header(self):
-        return "value"
+    value_column_header = "value"
 
     def connect_spine_db_editor(self, spine_db_editor):
         super().connect_spine_db_editor(spine_db_editor)
@@ -301,16 +324,45 @@ class ParameterValueTableView(ParameterTableView):
         if db_map is None:
             return None
         db_item = self.model().db_item(index)
+        if db_item is None:
+            return None
         return (db_map.db_url, {f: db_item[f] for f in self._pk_fields})
 
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class"""
+        raise NotImplementedError()
+
 
 class ObjectParameterDefinitionTableView(ObjectParameterTableMixin, ParameterDefinitionTableView):
     """A custom QTableView for the object parameter_definition pane in Spine db editor."""
 
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class"""
+        header_sections = [
+            ParameterTableHeaderSection(label) for label in ("database", "object_class_name", "parameter_name")
+        ]
+        return plot_parameter_table_selection(
+            self.model(), selection, header_sections, self.value_column_header, plot_widget
+        )
+
 
 class RelationshipParameterDefinitionTableView(RelationshipParameterTableMixin, ParameterDefinitionTableView):
     """A custom QTableView for the relationship parameter_definition pane in Spine db editor."""
 
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class"""
+        header_sections = [
+            ParameterTableHeaderSection(label)
+            for label in ("database", "relationship_class_name", "object_class_name_list", "parameter_name")
+        ]
+        for i, section in enumerate(header_sections):
+            if section.label == "object_class_name_list":
+                header_sections[i] = replace(section, separator=DB_ITEM_SEPARATOR)
+                break
+        return plot_parameter_table_selection(
+            self.model(), selection, header_sections, self.value_column_header, plot_widget
+        )
+
 
 class ObjectParameterValueTableView(ObjectParameterTableMixin, ParameterValueTableView):
     """A custom QTableView for the object parameter_value pane in Spine db editor."""
@@ -321,7 +373,14 @@ class ObjectParameterValueTableView(ObjectParameterTableMixin, ParameterValueTab
 
     @property
     def _pk_fields(self):
-        return ("object_class_name", "object_name", "parameter_name", "alternative_name")
+        return "object_class_name", "object_name", "parameter_name", "alternative_name"
+
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class."""
+        header_sections = [ParameterTableHeaderSection(label) for label in ("database",) + self._pk_fields]
+        return plot_parameter_table_selection(
+            self.model(), selection, header_sections, self.value_column_header, plot_widget
+        )
 
 
 class RelationshipParameterValueTableView(RelationshipParameterTableMixin, ParameterValueTableView):
@@ -334,7 +393,18 @@ class RelationshipParameterValueTableView(RelationshipParameterTableMixin, Param
 
     @property
     def _pk_fields(self):
-        return ("relationship_class_name", "object_name_list", "parameter_name", "alternative_name")
+        return "relationship_class_name", "object_name_list", "parameter_name", "alternative_name"
+
+    def _plot_selection(self, selection, plot_widget=None):
+        """See base class."""
+        header_sections = [ParameterTableHeaderSection(label) for label in ("database",) + self._pk_fields]
+        for i, section in enumerate(header_sections):
+            if section.label == "object_name_list":
+                header_sections[i] = replace(section, separator=DB_ITEM_SEPARATOR)
+                break
+        return plot_parameter_table_selection(
+            self.model(), selection, header_sections, self.value_column_header, plot_widget
+        )
 
 
 class PivotTableView(CopyPasteTableView):
@@ -352,12 +422,26 @@ class PivotTableView(CopyPasteTableView):
             parent (QWidget, optional): parent widget
         """
         super().__init__(parent)
+        self.setHorizontalScrollMode(QTableView.ScrollPerPixel)
+        self.setVerticalScrollMode(QTableView.ScrollPerPixel)
+        # NOTE: order of creation of header tables is important for them to stack properly
+        self._left_header_table = CopyPasteTableView(self)
+        self._top_header_table = CopyPasteTableView(self)
+        self._top_left_header_table = CopyPasteTableView(self)
+        self._left_header_table.setObjectName("left")
+        self._top_header_table.setObjectName("top")
+        self._top_left_header_table.setObjectName("top-left")
         self._spine_db_editor = None
         self._context = None
         self._fetch_more_timer = QTimer(self)
         self._fetch_more_timer.setSingleShot(True)
         self._fetch_more_timer.setInterval(100)
         self._fetch_more_timer.timeout.connect(self._fetch_more_visible)
+        self._left_header_table.verticalScrollBar().valueChanged.connect(self.verticalScrollBar().setValue)
+        self.verticalScrollBar().valueChanged.connect(self._left_header_table.verticalScrollBar().setValue)
+        self._top_header_table.horizontalScrollBar().valueChanged.connect(self.horizontalScrollBar().setValue)
+        self.horizontalScrollBar().valueChanged.connect(self._top_header_table.horizontalScrollBar().setValue)
+        self._init_header_tables()
 
     class _ContextBase:
         """Base class for pivot table view's contexts."""
@@ -367,6 +451,7 @@ class PivotTableView(CopyPasteTableView):
         _REMOVE_PARAMETER = "Remove parameter definitions"
         _REMOVE_ALTERNATIVE = "Remove alternatives"
         _REMOVE_SCENARIO = "Remove scenarios"
+        _DUPLICATE_SCENARIO = "Duplicate scenario"
 
         def __init__(self, view, db_editor, horizontal_header, vertical_header):
             """
@@ -400,9 +485,8 @@ class PivotTableView(CopyPasteTableView):
         def _refresh_selected_indexes(self):
             """Caches selected index lists."""
             self._clear_selection_lists()
-            source_model = self._view.source_model
             for index in map(self._view.model().mapToSource, self._view.selectedIndexes()):
-                self._to_selection_lists(index, source_model)
+                self._to_selection_lists(index)
 
         def remove_alternatives(self):
             """Removes selected alternatives from the database."""
@@ -420,16 +504,15 @@ class PivotTableView(CopyPasteTableView):
             self._update_actions_availability()
             self._menu.exec_(position)
 
-        def _to_selection_lists(self, index, source_model):
+        def _to_selection_lists(self, index):
             """Caches given index to corresponding selected index list.
 
             Args:
                 index (QModelIndex): index to cache
-                source_model (PivotTableModelBase): underlying model
             """
-            if source_model.index_in_headers(index):
-                top_left_id = source_model.top_left_id(index)
-                header_type = source_model.top_left_headers[top_left_id].header_type
+            if self._view.source_model.index_in_headers(index):
+                top_left_id = self._view.source_model.top_left_id(index)
+                header_type = self._view.source_model.top_left_headers[top_left_id].header_type
                 try:
                     self._header_selection_lists[header_type].append(index)
                 except KeyError:
@@ -540,8 +623,8 @@ class PivotTableView(CopyPasteTableView):
             self._plot_in_window_menu = self._menu.addMenu("Plot in window")
             self._plot_in_window_menu.triggered.connect(self._plot_in_window)
             self._menu.addSeparator()
-            self._menu.addAction(self._db_editor.ui.actionCopy)
-            self._menu.addAction(self._db_editor.ui.actionPaste)
+            self._menu.addAction(self._view.copy_action)
+            self._menu.addAction(self._view.paste_action)
             self._menu.addSeparator()
             self._remove_values_action = self._menu.addAction("Remove parameter values", self.remove_values)
             self._remove_objects_action = self._menu.addAction(self._REMOVE_OBJECT, self.remove_objects)
@@ -550,7 +633,6 @@ class PivotTableView(CopyPasteTableView):
             )
             self._remove_parameters_action = self._menu.addAction(self._REMOVE_PARAMETER, self.remove_parameters)
             self._remove_alternatives_action = self._menu.addAction(self._REMOVE_ALTERNATIVE, self.remove_alternatives)
-            self._menu.aboutToShow.connect(self._db_editor.refresh_copy_paste_actions)
 
         def open_in_editor(self):
             """Opens the parameter value editor for the first selected cell."""
@@ -560,17 +642,18 @@ class PivotTableView(CopyPasteTableView):
         def plot(self):
             """Plots the selected cells."""
             selected_indexes = self._view.selectedIndexes()
-            hints = PivotTablePlottingHints()
             model = self._view.model()
             try:
-                plot_window = plot_selection(model, selected_indexes, hints)
+                plot_window = plot_pivot_table_selection(model, selected_indexes)
             except PlottingError as error:
-                report_plotting_failure(error, self)
+                report_plotting_failure(error, self._view)
                 return
+            source_model = model.sourceModel()
             plotted_column_names = {
-                hints.column_label(model, index.column())
+                source_model.column_name(index.column())
                 for index in selected_indexes
-                if hints.is_index_in_data(model, index)
+                if source_model.index_in_data(model.mapToSource(index))
+                or source_model.column_is_index_column(model.mapToSource(index).column())
             }
             plot_window.use_as_window(self._view, ", ".join(plotted_column_names))
             plot_window.show()
@@ -584,12 +667,11 @@ class PivotTableView(CopyPasteTableView):
                 self.plot()
                 return
             selected_indexes = self._view.selectedIndexes()
-            hints = PivotTablePlottingHints()
             try:
-                plot_selection(self._view.model(), selected_indexes, hints, plot_window)
+                plot_pivot_table_selection(self._view.model(), selected_indexes, plot_window)
                 plot_window.raise_()
             except PlottingError as error:
-                report_plotting_failure(error, self)
+                report_plotting_failure(error, self._view)
 
         def remove_parameters(self):
             """Removes selected parameter definitions from the database."""
@@ -619,17 +701,17 @@ class PivotTableView(CopyPasteTableView):
                 db_map_typed_data.setdefault(db_map, {}).setdefault("parameter_value", set()).add(id_)
             self._db_editor.db_mngr.remove_items(db_map_typed_data)
 
-        def _show_context_menu(self, position):
-            """See base class."""
-            _prepare_plot_in_window_menu(self._plot_in_window_menu)
+        @Slot(QPoint)
+        def show_context_menu(self, position):
+            prepare_plot_in_window_menu(self._plot_in_window_menu)
             super().show_context_menu(position)
 
-        def _to_selection_lists(self, index, source_model):
+        def _to_selection_lists(self, index):
             """See base class."""
-            if source_model.index_in_data(index):
+            if self._view.source_model.index_in_data(index):
                 self._selected_value_indexes.append(index)
             else:
-                super()._to_selection_lists(index, source_model)
+                super()._to_selection_lists(index)
 
         def _update_actions_availability(self):
             """See base class."""
@@ -662,11 +744,10 @@ class PivotTableView(CopyPasteTableView):
 
         def populate_context_menu(self):
             """See base class."""
-            self._menu.addAction(self._db_editor.ui.actionCopy)
-            self._menu.addAction(self._db_editor.ui.actionPaste)
+            self._menu.addAction(self._view.copy_action)
+            self._menu.addAction(self._view.paste_action)
             self._menu.addSeparator()
             self._remove_objects_action = self._menu.addAction(self._REMOVE_OBJECT, self.remove_objects)
-            self._menu.aboutToShow.connect(self._db_editor.refresh_copy_paste_actions)
 
         def _update_actions_availability(self):
             """See base class."""
@@ -687,6 +768,7 @@ class PivotTableView(CopyPasteTableView):
             self._toggle_alternatives_checked = QAction("Check/uncheck selected")
             self._toggle_alternatives_checked.triggered.connect(self._toggle_checked_state)
             self._remove_scenarios_action = None
+            self._duplicate_scenario_action = None
             horizontal_header = ScenarioAlternativePivotHeaderView(Qt.Horizontal, "columns", view)
             horizontal_header.context_menu_requested.connect(self.show_context_menu)
             vertical_header = ScenarioAlternativePivotHeaderView(Qt.Vertical, "rows", view)
@@ -696,8 +778,8 @@ class PivotTableView(CopyPasteTableView):
 
         def _clear_selection_lists(self):
             """See base class."""
-            self._selected_scenario_indexes = list()
-            self._selected_scenario_alternative_indexes = list()
+            self._selected_scenario_indexes.clear()
+            self._selected_scenario_alternative_indexes.clear()
             super()._clear_selection_lists()
 
         def populate_context_menu(self):
@@ -708,12 +790,12 @@ class PivotTableView(CopyPasteTableView):
             self._menu.addSeparator()
             self._menu.addAction(self._toggle_alternatives_checked)
             self._menu.addSeparator()
-            self._menu.addAction(self._db_editor.ui.actionCopy)
-            self._menu.addAction(self._db_editor.ui.actionPaste)
+            self._menu.addAction(self._view.copy_action)
+            self._menu.addAction(self._view.paste_action)
             self._menu.addSeparator()
             self._remove_alternatives_action = self._menu.addAction(self._REMOVE_ALTERNATIVE, self.remove_alternatives)
             self._remove_scenarios_action = self._menu.addAction(self._REMOVE_SCENARIO, self.remove_scenarios)
-            self._menu.aboutToShow.connect(self._db_editor.refresh_copy_paste_actions)
+            self._duplicate_scenario_action = self._menu.addAction(self._DUPLICATE_SCENARIO, self.duplicate_scenario)
 
         def remove_scenarios(self):
             """Removes selected scenarios from the database."""
@@ -724,18 +806,25 @@ class PivotTableView(CopyPasteTableView):
                 db_map_typed_data.setdefault(db_map, {}).setdefault("scenario", set()).add(id_)
             self._db_editor.db_mngr.remove_items(db_map_typed_data)
 
-        def _to_selection_lists(self, index, source_model):
+        def duplicate_scenario(self):
+            """Duplicates current scenario in the database."""
+            index = self._selected_scenario_indexes[0]
+            db_map, scen_id = self._view.source_model._header_id(index)
+            self._db_editor.duplicate_scenario(db_map, scen_id)
+
+        def _to_selection_lists(self, index):
             """See base class."""
-            if source_model.index_in_data(index):
+            if self._view.source_model.index_in_data(index):
                 self._selected_scenario_alternative_indexes.append(index)
             else:
-                super()._to_selection_lists(index, source_model)
+                super()._to_selection_lists(index)
 
         def _update_actions_availability(self):
             """See base class."""
             self._generate_scenarios_action.setEnabled(bool(self._selected_alternative_indexes))
             self._remove_alternatives_action.setEnabled(bool(self._selected_alternative_indexes))
             self._remove_scenarios_action.setEnabled(bool(self._selected_scenario_indexes))
+            self._duplicate_scenario_action.setEnabled(len(self._selected_scenario_indexes) == 1)
 
         def _open_scenario_generator(self):
             """Opens the scenario generator dialog."""
@@ -773,7 +862,9 @@ class PivotTableView(CopyPasteTableView):
 
     def connect_spine_db_editor(self, spine_db_editor):
         self._spine_db_editor = spine_db_editor
+        self.set_external_copy_and_paste_actions(spine_db_editor.ui.actionCopy, spine_db_editor.ui.actionPaste)
         self._spine_db_editor.pivot_table_proxy.sourceModelChanged.connect(self._change_context)
+        self.selectionModel().selectionChanged.connect(self._refresh_copy_paste_actions)
 
     @Slot()
     def _change_context(self):
@@ -811,8 +902,31 @@ class PivotTableView(CopyPasteTableView):
         old_model = self.model()
         if old_model:
             old_model.model_data_changed.disconnect(self._fetch_more_timer.start)
+            old_model.modelReset.disconnect(self._update_header_tables)
         super().setModel(model)
+        for header_table in (self._left_header_table, self._top_header_table, self._top_left_header_table):
+            header_table.setModel(model)
         model.model_data_changed.connect(self._fetch_more_timer.start)
+        model.modelReset.connect(self._update_header_tables)
+
+    def setIndexWidget(self, proxy_index, widget):
+        self._top_left_header_table.setIndexWidget(proxy_index, widget)
+
+    def setHorizontalHeader(self, horizontal_header):
+        super().setHorizontalHeader(horizontal_header)
+        horizontal_header.sectionResized.connect(self._update_section_width)
+        for header_table in (self._left_header_table, self._top_header_table, self._top_left_header_table):
+            header_table.horizontalHeader().setResizeContentsPrecision(horizontal_header.resizeContentsPrecision())
+
+    def setVerticalHeader(self, vertical_header):
+        super().setVerticalHeader(vertical_header)
+        vertical_header.sectionResized.connect(self._update_section_height)
+        for header_table in (self._left_header_table, self._top_header_table, self._top_left_header_table):
+            header_table.verticalHeader().setDefaultSectionSize(vertical_header.defaultSectionSize())
+
+    def resizeEvent(self, ev):
+        super().resizeEvent(ev)
+        self._update_header_tables_geometry()
 
     def _fetch_more_visible(self):
         model = self.model()
@@ -821,6 +935,67 @@ class PivotTableView(CopyPasteTableView):
         if scrollbar_at_max and model.canFetchMore(QModelIndex()):
             model.fetchMore(QModelIndex())
 
+    def _init_header_tables(self):
+        # NOTE: order of the iteration below is important for calls to stackUnder
+        for header_table in (self._top_left_header_table, self._top_header_table, self._left_header_table):
+            header_table.setFocusPolicy(Qt.NoFocus)
+            header_table.setStyleSheet(self.styleSheet())
+            header_table.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
+            header_table.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
+            header_table.show()
+            header_table.verticalHeader().hide()
+            header_table.horizontalHeader().hide()
+            header_table.setHorizontalScrollMode(QTableView.ScrollPerPixel)
+            header_table.setVerticalScrollMode(QTableView.ScrollPerPixel)
+            header_table.setStyleSheet("QTableView { border: none;}")
+            self.viewport().stackUnder(header_table)
+
+    def _update_header_tables(self):
+        # Top
+        for header_table in (self._top_header_table, self._top_left_header_table):
+            for i in range(0, self.source_model.headerRowCount()):
+                header_table.setRowHeight(i, self.rowHeight(i))
+                header_table.setRowHidden(i, False)
+            for i in range(self.source_model.headerRowCount(), self.source_model.rowCount() - 1):
+                header_table.setRowHidden(i, True)
+        # Left
+        for header_table in (self._left_header_table, self._top_left_header_table):
+            for j in range(0, self.source_model.headerColumnCount()):
+                header_table.setColumnWidth(j, self.columnWidth(j))
+                header_table.setColumnHidden(j, False)
+            for j in range(self.source_model.headerColumnCount(), self.source_model.columnCount() - 1):
+                header_table.setColumnHidden(j, True)
+        self._update_header_tables_geometry()
+
+    @Slot(int, int, int)
+    def _update_section_width(self, logical_index, _old_size, new_size):
+        for header_table in (self._left_header_table, self._top_header_table, self._top_left_header_table):
+            header_table.setColumnWidth(logical_index, new_size)
+        self._update_header_tables_geometry()
+
+    @Slot(int, int, int)
+    def _update_section_height(self, logical_index, _old_size, new_size):
+        for header_table in (self._left_header_table, self._top_header_table, self._top_left_header_table):
+            header_table.setRowHeight(logical_index, new_size)
+        self._update_header_tables_geometry()
+
+    def _update_header_tables_geometry(self):
+        if not self.source_model:
+            return
+        x = self.verticalHeader().width() + self.frameWidth()
+        y = self.horizontalHeader().height() + self.frameWidth()
+        header_w = sum(self.columnWidth(j) for j in range(0, self.source_model.headerColumnCount()))
+        header_h = sum(self.rowHeight(i) for i in range(0, self.source_model.headerRowCount()))
+        total_w = self.viewport().width()
+        total_h = self.viewport().height()
+        self._left_header_table.setGeometry(x, y, header_w, total_h)
+        self._top_header_table.setGeometry(x, y, total_w, header_h)
+        self._top_left_header_table.setGeometry(x, y, header_w, header_h)
+
+    @Slot(QModelIndex, QModelIndex)
+    def _refresh_copy_paste_actions(self, _, __):
+        self._spine_db_editor.refresh_copy_paste_actions()
+
 
 class FrozenTableView(QTableView):
 
@@ -853,6 +1028,7 @@ class MetadataTableViewBase(CopyPasteTableView):
         super().__init__(parent)
         self.verticalHeader().setDefaultSectionSize(preferred_row_height(self))
         self._menu = QMenu(self)
+        self._db_editor = None
 
     def connect_spine_db_editor(self, db_editor):
         """Finishes view's initialization.
@@ -860,8 +1036,11 @@ class MetadataTableViewBase(CopyPasteTableView):
         Args:
              db_editor (SpineDBEditor): database editor instance
         """
-        self._populate_context_menu(db_editor)
+        self._db_editor = db_editor
+        self.set_external_copy_and_paste_actions(db_editor.ui.actionCopy, db_editor.ui.actionPaste)
+        self._populate_context_menu()
         self._enable_delegates(db_editor)
+        self.selectionModel().selectionChanged.connect(self._refresh_copy_paste_actions)
 
     def contextMenuEvent(self, event):
         menu_position = event.globalPos()
@@ -884,14 +1063,10 @@ class MetadataTableViewBase(CopyPasteTableView):
             db_editor (SpineDBEditor): database editor
         """
 
-    def _populate_context_menu(self, db_editor):
-        """Fills context menu with actions.
-
-        Args:
-            db_editor (SpineDBEditor): database editor
-        """
-        self._menu.addAction(db_editor.ui.actionCopy)
-        self._menu.addAction(db_editor.ui.actionPaste)
+    def _populate_context_menu(self,):
+        """Fills context menu with actions."""
+        self._menu.addAction(self.copy_action)
+        self._menu.addAction(self.paste_action)
         self._menu.addSeparator()
         self._menu.addAction("Remove row(s)", self._remove_selected)
 
@@ -905,6 +1080,10 @@ class MetadataTableViewBase(CopyPasteTableView):
         """
         self.model().setData(index, value)
 
+    @Slot(QModelIndex, QModelIndex)
+    def _refresh_copy_paste_actions(self):
+        self._db_editor.refresh_copy_paste_actions()
+
 
 class MetadataTableView(MetadataTableViewBase):
     """Table view for metadata."""
