@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -106,7 +106,6 @@ class EntityTreeView(CopyTreeView):
     def connect_signals(self):
         """Connects signals."""
         self.selectionModel().selectionChanged.connect(self._handle_selection_changed)
-        self._menu.aboutToShow.connect(self._spine_db_editor.refresh_copy_paste_actions)
 
     def rowsInserted(self, parent, start, end):
         super().rowsInserted(parent, start, end)
@@ -140,6 +139,7 @@ class EntityTreeView(CopyTreeView):
     @Slot(QItemSelection, QItemSelection)
     def _handle_selection_changed(self, selected, deselected):
         """Classifies selection by item type and emits signal."""
+        self._spine_db_editor.refresh_copy_paste_actions()
         self._refresh_selected_indexes()
         if not self.selectionModel().hasSelection():
             return
@@ -388,7 +388,7 @@ class ItemTreeView(CopyTreeView):
 
     def connect_signals(self):
         """Connects signals."""
-        self._menu.aboutToShow.connect(self._spine_db_editor.refresh_copy_paste_actions)
+        self.selectionModel().selectionChanged.connect(self._refresh_copy_paste_actions)
 
     def remove_selected(self):
         """Removes items selected in the view."""
@@ -421,6 +421,11 @@ class ItemTreeView(CopyTreeView):
         self.update_actions_availability(item)
         self._menu.exec_(event.globalPos())
 
+    @Slot(QModelIndex, QModelIndex)
+    def _refresh_copy_paste_actions(self, _, __):
+        """Refreshes copy and paste actions enabled state."""
+        self._spine_db_editor.refresh_copy_paste_actions()
+
 
 class ToolFeatureTreeView(ItemTreeView):
     """Custom QTreeView class for tools and features in SpineDBEditor."""
