@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -190,9 +190,12 @@ class TabularViewMixin:
         self.current_class_id = class_id
         self.current_class_type = current_class_item.item_type
         self.current_class_name = current_class_item.display_data
+        self.clear_pivot_table()
 
     @staticmethod
     def _get_current_class_item(current_index):
+        if not current_index.isValid():
+            return None
         item = current_index.model().item_from_index(current_index)
         while item.item_type != "root":
             if item.item_type in ("object_class", "relationship_class"):
@@ -501,7 +504,6 @@ class TabularViewMixin:
         """Reloads pivot table."""
         if not self._can_build_pivot_table():
             return
-        self.clear_pivot_table()
         self.pivot_table_model = {
             self._PARAMETER_VALUE: ParameterValuePivotTableModel,
             self._RELATIONSHIP: RelationshipPivotTableModel,
