@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -226,7 +226,7 @@ class SpineDBManager(QObject):
         for item_type, signal in self.updated_signals.items():
             signal.connect(lambda db_map_data, item_type=item_type: self.cache_items(item_type, db_map_data))
         # Uncache
-        self.items_removed.connect(self.uncache_items)
+        self.items_removed.connect(self.uncache_removed_items)
         # Icons
         self.object_classes_added.connect(self.update_icons)
         self.object_classes_updated.connect(self.update_icons)
@@ -286,9 +286,10 @@ class SpineDBManager(QObject):
         if db_map.connection.closed:
             return False
         try:
-            return self._get_worker(db_map).can_fetch_more(parent)
+            worker = self._get_worker(db_map)
         except KeyError:
             return False
+        return worker.can_fetch_more(parent)
 
     def fetch_more(self, db_map, parent):
         """Fetches more items of given type from given db.
@@ -299,7 +300,11 @@ class SpineDBManager(QObject):
         """
         if db_map.connection.closed:
             return
-        self._get_worker(db_map).fetch_more(parent)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return
+        worker.fetch_more(parent)
 
     def cache_items(self, item_type, db_map_data):
         """Caches data for a given type.
@@ -323,8 +328,8 @@ class SpineDBManager(QObject):
     def _pop_item(self, db_map, item_type, id_):
         return self._cache.get(db_map, {}).get(item_type, {}).pop(id_, {})
 
-    def uncache_items(self, db_map_typed_ids):
-        """Removes data from cache.
+    def uncache_removed_items(self, db_map_typed_ids):
+        """Removes data that has been removed from the database also from cache.
 
         Args:
             db_map_typed_ids (dict): items to remove
@@ -337,8 +342,8 @@ class SpineDBManager(QObject):
                 items = [item for item in (self._pop_item(db_map, item_type, id_) for id_ in ids) if item]
                 if items:
                     db_map_data[db_map] = items
-                    worker = self._get_worker(db_map)
-                    worker.reset_queries(item_type)
+                    if db_map not in self._workers:
+                        continue
             if db_map_data:
                 typed_db_map_data[item_type] = db_map_data
                 signal.emit(db_map_data)
@@ -347,7 +352,10 @@ class SpineDBManager(QObject):
 
     @busy_effect
     def get_db_map_cache(self, db_map, item_types=None, only_descendants=False, include_ancestors=False):
-        worker = self._get_worker(db_map)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return {}
         worker.fetch_all(item_types=item_types, only_descendants=only_descendants, include_ancestors=include_ancestors)
         return self._cache.setdefault(db_map, {})
 
@@ -435,6 +443,7 @@ class SpineDBManager(QObject):
             if worker is not None:
                 worker.close_db_map()
                 worker.clean_up()
+            self._cache.pop(db_map, None)
             del self.undo_stack[db_map]
             del self.undo_action[db_map]
             del self.redo_action[db_map]
@@ -512,9 +521,11 @@ class SpineDBManager(QObject):
                 db_map.codename = codename
             return db_map
         worker = SpineDBWorker(self, url)
-        db_map, err = worker.get_db_map(codename=codename, upgrade=upgrade, create=create)
-        if err is not None:
-            raise err
+        try:
+            db_map = worker.get_db_map(codename=codename, upgrade=upgrade, create=create)
+        except Exception as error:
+            worker.clean_up()
+            raise error
         self._workers[db_map] = worker
         self._db_maps[url] = db_map
         self.db_map_locks[db_map] = QMutex(QMutex.Recursive)
@@ -543,17 +554,16 @@ class SpineDBManager(QObject):
             except AttributeError:
                 pass
 
-    def unregister_listener(self, listener, commit_dirty, commit_msg, *db_maps):
+    def unregister_listener(self, listener, *db_maps, commit_dirty=False, commit_msg=""):
         """Unregisters given listener from given db_map signals.
         If any of the db_maps becomes an orphan and is dirty, prompts user to commit or rollback.
 
         Args:
             listener (object)
+            *db_maps (DiffDatabaseMapping)
             commit_dirty (bool): True to commit dirty database mapping, False to roll back
             commit_msg (str): commit message
-            *db_maps (DiffDatabaseMapping)
         """
-        dirty_orphan_db_maps = self.dirty_or_orphan(listener, *db_maps)
         for db_map in db_maps:
             self.signaller.remove_db_map_listener(db_map, listener)
             try:
@@ -561,17 +571,29 @@ class SpineDBManager(QObject):
                 self.undo_stack[db_map].cleanChanged.disconnect(listener.update_commit_enabled)
             except AttributeError:
                 pass
-        if dirty_orphan_db_maps:
+        dirty_db_maps = self.dirty_and_without_editors(listener, *db_maps)
+        if dirty_db_maps:
             if commit_dirty:
-                self.commit_session(commit_msg, *dirty_orphan_db_maps)
+                self.commit_session(commit_msg, *dirty_db_maps)
             else:
-                self.rollback_session(*dirty_orphan_db_maps)
+                self.rollback_session(*dirty_db_maps)
         for db_map in db_maps:
             if not self.signaller.db_map_listeners(db_map):
                 self.close_session(db_map.db_url)
 
+    def is_dirty(self, db_map):
+        """Returns True if mapping has pending changes.
+
+        Args:
+            db_map (DiffDatabaseMapping): database mapping
+
+        Returns:
+            bool: True if db_map has pending changes, False otherwise
+        """
+        return not self.undo_stack[db_map].isClean()
+
     def dirty(self, *db_maps):
-        """Checks which of the given database mappings are dirty.
+        """Filters clean mappings from given database maps.
 
         Args:
             *db_maps: mappings to check
@@ -579,26 +601,23 @@ class SpineDBManager(QObject):
         Return:
             list of DiffDatabaseMapping: dirty  mappings
         """
-        return [db_map for db_map in db_maps if not self.undo_stack[db_map].isClean() or db_map.has_pending_changes()]
+        return [db_map for db_map in db_maps if self.is_dirty(db_map)]
 
-    def dirty_or_orphan(self, listener, *db_maps):
-        """Checks which of the given database mappings are dirty or orphaned.
+    def dirty_and_without_editors(self, listener, *db_maps):
+        """Checks which of the given database mappings are dirty and have no editors.
 
         Args:
             listener (Any): a listener object
             *db_maps: mappings to check
 
         Return:
-            list of DiffDatabaseMapping: dirty or orphaned mappings
+            list of DiffDatabaseMapping: mappings that are dirty and don't have editors
         """
 
-        def is_dirty(db_map):
-            return not self.undo_stack[db_map].isClean() or db_map.has_pending_changes()
+        def has_no_editors(db_map):
+            return not set(self.signaller.db_map_editors(db_map)) - {listener}
 
-        def is_orphan(db_map):
-            return not set(self.signaller.db_map_listeners(db_map)) - {listener}
-
-        return [db_map for db_map in db_maps if is_orphan(db_map) and is_dirty(db_map)]
+        return [db_map for db_map in db_maps if has_no_editors(db_map) and self.is_dirty(db_map)]
 
     def clean_up(self):
         while self._workers:
@@ -623,7 +642,21 @@ class SpineDBManager(QObject):
             cookie (object, optional): a free form identifier which will be forwarded to ``session_committed`` signal
         """
         for db_map in dirty_db_maps:
-            self._get_worker(db_map).commit_session(commit_msg, cookie)
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                continue
+            worker.commit_session(commit_msg, cookie)
+
+    def notify_session_committed(self, cookie, *db_maps):
+        """Notifies manager and listeners when a commit has taken place by a third party.
+
+        Args:
+            cookie (Any): commit cookie
+            *db_maps: database maps that were committed
+        """
+        self.refresh_session(*db_maps)
+        self.session_committed.emit(set(db_maps), cookie)
 
     def rollback_session(self, *dirty_db_maps):
         """
@@ -633,14 +666,22 @@ class SpineDBManager(QObject):
             *dirty_db_maps: dirty database maps to commit
         """
         for db_map in dirty_db_maps:
-            self._get_worker(db_map).rollback_session()
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                continue
+            worker.rollback_session()
         self._restart_fetching(dirty_db_maps)
 
     def _restart_fetching(self, db_maps):
         """Restarts fetching"""
         for db_map in db_maps:
             del self._cache[db_map]
-            self._get_worker(db_map).reset_queries()
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                continue
+            worker.reset_queries()
 
     def entity_class_renderer(self, db_map, entity_type, entity_class_id, for_group=False):
         """Returns an icon renderer for a given entity class.
@@ -697,7 +738,10 @@ class SpineDBManager(QObject):
         item = self._cache.get(db_map, {}).get(item_type, {}).get(id_, {})
         if only_visible and item:
             return item
-        worker = self._get_worker(db_map)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return {}
         worker.fetch_all(item_types={item_type})
         return self._cache.get(db_map, {}).get(item_type, {}).get(id_, {})
 
@@ -718,7 +762,10 @@ class SpineDBManager(QObject):
         items = list(self._cache.get(db_map, {}).get(item_type, {}).values())
         if only_visible:
             return items
-        worker = self._get_worker(db_map)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return []
         worker.fetch_all(item_types={item_type})
         return list(self._cache.get(db_map, {}).get(item_type, {}).values())
 
@@ -861,8 +908,8 @@ class SpineDBManager(QObject):
             return self.tool_tip_data_from_parsed(parsed_value)
         if role == Qt.TextAlignmentRole:
             if isinstance(parsed_value, str):
-                return Qt.AlignLeft
-            return Qt.AlignRight
+                return int(Qt.AlignLeft | Qt.AlignVCenter)
+            return int(Qt.AlignRight | Qt.AlignVCenter)
         if role == PARSED_ROLE:
             return parsed_value
         return None
@@ -1000,13 +1047,22 @@ class SpineDBManager(QObject):
 
     def _add_or_update_items(self, db_map_data, method_name, item_type, signal_name, check):
         for db_map, data in db_map_data.items():
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                # We're closing the kiosk.
+                continue
             cache = self.get_db_map_cache(db_map, {item_type}, include_ancestors=True)
-            self._get_worker(db_map).add_or_update_items(data, method_name, item_type, signal_name, check, cache)
+            worker.add_or_update_items(data, method_name, item_type, signal_name, check, cache)
 
     def _readd_items(self, db_map_data, method_name, item_type, signal_name):
         for db_map, data in db_map_data.items():
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                continue
             cache = self.get_db_map_cache(db_map, {item_type}, include_ancestors=True)
-            self._get_worker(db_map).readd_items(data, method_name, item_type, signal_name, cache)
+            worker.readd_items(data, method_name, item_type, signal_name, cache)
 
     def add_alternatives(self, db_map_data):
         """Adds alternatives to db.
@@ -1394,7 +1450,23 @@ class SpineDBManager(QObject):
                 macro.setObsolete(True)
                 self.undo_stack[db_map].undo()
 
+    def purge_items(self, db_map_purgable_items):
+        """Purges selected items from given database.
+
+        Args:
+            db_map_purgable_items (dict): mapping from database map to list of purgable item types
+        """
+        db_map_typed_data = {
+            db_map: {
+                item_type: {x["id"] for x in self.get_items(db_map, item_type, only_visible=False)}
+                for item_type in item_types
+            }
+            for db_map, item_types in db_map_purgable_items.items()
+        }
+        self.remove_items(db_map_typed_data)
+
     def remove_items(self, db_map_typed_ids):
+        """Pushes a command to remove items to undo stack."""
         for db_map, ids_per_type in db_map_typed_ids.items():
             self.undo_stack[db_map].push(RemoveItemsCommand(self, db_map, ids_per_type))
 
@@ -1406,7 +1478,11 @@ class SpineDBManager(QObject):
             db_map_typed_ids (dict): mapping DiffDatabaseMapping to item type (str) to lists of items to remove
         """
         for db_map, ids_per_type in db_map_typed_ids.items():
-            self._get_worker(db_map).remove_items(ids_per_type)
+            try:
+                worker = self._get_worker(db_map)
+            except KeyError:
+                continue
+            worker.remove_items(ids_per_type)
 
     @staticmethod
     def db_map_ids(db_map_data):
@@ -1669,6 +1745,15 @@ class SpineDBManager(QObject):
                     db_map_cascading_data.setdefault(db_map, []).append(item)
         return db_map_cascading_data
 
+    def find_cascading_scenario_alternatives_by_scenario(self, db_map_ids, only_visible=True):
+        """Finds and returns cascading scenario alternatives for the given scenario ids."""
+        db_map_cascading_data = dict()
+        for db_map, ids in db_map_ids.items():
+            for item in self.get_items(db_map, "scenario_alternative", only_visible=only_visible):
+                if item["scenario_id"] in ids:
+                    db_map_cascading_data.setdefault(db_map, []).append(item)
+        return db_map_cascading_data
+
     def _refresh_parameter_value_lists(self, db_map_data):
         """Refreshes cached parameter value lists when updating list values.
 
@@ -1865,7 +1950,19 @@ class SpineDBManager(QObject):
         db_map_cascading_data = self.find_cascading_parameter_definitions_by_list_value(self.db_map_ids(db_map_data))
         self.parameter_definitions_updated.emit(db_map_cascading_data)
 
-    def duplicate_object(self, db_maps, object_data, orig_name, dup_name):
+    def duplicate_scenario(self, scen_data, dup_name, db_map):
+        data = self._get_data_for_export(scen_data)
+        data = {
+            "scenarios": [(dup_name, active, description) for (_, active, description) in data.get("scenarios", [])],
+            "alternatives": data.get("alternatives", []),
+            "scenario_alternatives": [
+                (dup_name, alt_name, before_alt_name)
+                for (_, alt_name, before_alt_name) in data.get("scenario_alternatives", [])
+            ],
+        }
+        self.import_data({db_map: data}, command_text="Duplicate scenario")
+
+    def duplicate_object(self, object_data, orig_name, dup_name, db_maps):
         _replace_name = lambda name_list: [name if name != orig_name else dup_name for name in name_list]
         data = self._get_data_for_export(object_data)
         data = {
@@ -1973,7 +2070,11 @@ class SpineDBManager(QObject):
         Returns:
             list of namedtuple: entity metadata records
         """
-        return self._get_worker(db_map).get_entity_metadata(entity_id)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return []
+        return worker.get_entity_metadata(entity_id)
 
     def get_parameter_value_metadata(self, db_map, parameter_value_id):
         """Returns metadata records for given parameter value.
@@ -1985,10 +2086,17 @@ class SpineDBManager(QObject):
         Returns:
             list of namedtuple: parameter value metadata records
         """
-        return self._get_worker(db_map).get_parameter_value_metadata(parameter_value_id)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return []
+        return worker.get_parameter_value_metadata(parameter_value_id)
 
     def get_items_for_commit(self, db_map, commit_id):
-        worker = self._get_worker(db_map)
+        try:
+            worker = self._get_worker(db_map)
+        except KeyError:
+            return {}
         worker.fetch_all()
         return worker.commit_cache.get(commit_id, {})
 
