@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -43,6 +43,7 @@ class SpineDBParcel:
         return self._data
 
     def _get_fields(self, db_map, item_type, field, ids):
+        """Returns a list of field values for items of given type, having given ids."""
         if ids is Asterisk:
             fields = {x.get(field) for x in self.db_mngr.get_items(db_map, item_type, only_visible=False)}
         else:
@@ -264,6 +265,13 @@ class SpineDBParcel:
         db_map_ids = {db_map: ids - param_val_ids.get(db_map, set()) for db_map, ids in db_map_ids.items()}
         self.push_relationship_ids(db_map_ids)
 
+    def full_push_scenario_ids(self, db_map_ids):
+        self.push_scenario_ids(db_map_ids)
+        scenario_alternative_ids = self.db_mngr.db_map_ids(
+            self.db_mngr.find_cascading_scenario_alternatives_by_scenario(db_map_ids, only_visible=False)
+        )
+        self.push_scenario_alternative_ids(scenario_alternative_ids)
+
     def inner_push_object_ids(self, db_map_ids):
         """Pushes object ids, cascading relationship ids, and the associated parameter values,
         but not any entity classes or parameter definitions.
