@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -58,15 +58,35 @@ class SpineDBSignaller(QObject):
         self.listeners.setdefault(listener, set()).add(db_map)
 
     def remove_db_map_listener(self, db_map, listener):
-        """Removes db_map from the the maps listener listens to."""
+        """Removes db_map from the maps listener listens to."""
         db_maps = self.listeners.get(listener, set())
         db_maps.discard(db_map)
         if not db_maps:
-            self.listeners.pop(listener)
+            # We cannot just pop here because we might be iterating self.listeners
+            # in _call_in_listeners().
+            self.listeners = {l: db_maps for l, db_maps in self.listeners.items() if l is not listener}
 
     def db_map_listeners(self, db_map):
         return set(listener for listener, db_maps in self.listeners.items() if db_map in db_maps)
 
+    def db_map_editors(self, db_map):
+        """Creates a set of given database map's editors.
+
+        Editors are listeners that are interested whether the database is dirty.
+        They are expected to implement ``is_db_map_editor()`` which returns a bool.
+
+        Args:
+            db_map (DiffDatabaseMapping): database map
+
+        Returns:
+            set: database mapping's editors
+        """
+        return {
+            listener
+            for listener in self.db_map_listeners(db_map)
+            if getattr(listener, "is_db_map_editor", lambda: False)()
+        }
+
     def connect_signals(self):
         """Connects signals."""
         # Added
