@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -71,7 +71,7 @@ class DesignGraphicsScene(CustomGraphicsScene):
         self.link_drawer = None
         self.icon_group = set()  # Group of project item icons that are moving together
         self.dirty_links = set()
-        self._timer = QTimer()
+        self._timer = QTimer(self)
         self._timer.setInterval(5)
         self._timer.timeout.connect(self._handle_timeout)
         self._timer.start()
@@ -99,23 +99,35 @@ class DesignGraphicsScene(CustomGraphicsScene):
 
     def mousePressEvent(self, event):
         """Puts link drawer to sleep and log message if it looks like the user doesn't know what they're doing."""
-        was_drawing = self.link_drawer is not None
+        if (
+            self._toolbox.qsettings().value("appSettings/dragToDrawLinks", defaultValue="false") == "false"
+            and self._finish_link()
+        ):
+            return
         super().mousePressEvent(event)
-        if was_drawing and self.link_drawer is not None:
-            self.link_drawer.sleep()
-            if event.button() == Qt.LeftButton:
-                self.emit_connection_failed()
 
     def mouseReleaseEvent(self, event):
         """Makes link if drawer is released over a valid connector button."""
-        super().mouseReleaseEvent(event)
-        if self.link_drawer is None or self.link_drawer.src_connector.isUnderMouse():
+        if (
+            self._toolbox.qsettings().value("appSettings/dragToDrawLinks", defaultValue="false") == "true"
+            and self._finish_link()
+        ):
             return
+        super().mouseReleaseEvent(event)
+
+    def _finish_link(self):
+        if self.link_drawer is None:
+            return False
+        if self.link_drawer.src_connector.isUnderMouse():
+            self.link_drawer.sleep()
+            return False
         if self.link_drawer.dst_connector is None:
             self.link_drawer.sleep()
             self.emit_connection_failed()
-            return
+            return False
+        self.link_drawer.dst_connector.set_normal_brush()
         self.link_drawer.add_link()
+        return True
 
     def emit_connection_failed(self):
         self._toolbox.msg_warning.emit(
