@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -502,4 +502,4 @@ class DesignQGraphicsView(CustomQGraphicsView):
         if not event.isAccepted():
             event.accept()
             global_pos = self.viewport().mapToGlobal(event.pos())
-            self._toolbox.show_project_item_context_menu(global_pos, None)
+            self._toolbox.show_project_or_item_context_menu(global_pos, None)
