@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -25,7 +25,7 @@ from numbers import Number
 import re
 from PySide2.QtWidgets import QTableView, QApplication, QAction
 from PySide2.QtCore import Qt, Slot, QItemSelection, QItemSelectionModel, QPoint
-from PySide2.QtGui import QKeySequence
+from PySide2.QtGui import QKeySequence, QIcon
 from spinedb_api import (
     DateTime,
     Duration,
@@ -45,21 +45,55 @@ _ = csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))
 class CopyPasteTableView(QTableView):
     """Custom QTableView class with copy and paste methods."""
 
-    def keyPressEvent(self, event):
-        """Copies and pastes to and from clipboard in Excel-like format."""
-        if event.matches(QKeySequence.Copy):
-            if not self.copy():
-                super().keyPressEvent(event)
-        elif event.matches(QKeySequence.Paste):
-            if not self.paste():
-                super().keyPressEvent(event)
-        elif event.matches(QKeySequence.Delete):
-            if not self.delete_content():
-                super().keyPressEvent(event)
-        else:
-            super().keyPressEvent(event)
+    def __init__(self, parent=None):
+        super().__init__(parent)
+        self._copy_action = None
+        self._paste_action = None
+        self._delete_action = QAction("Delete", self)
+        self._delete_action.setShortcut(QKeySequence.Delete)
+        self.addAction(self._delete_action)
+        self._delete_action.triggered.connect(self.delete_content)
 
-    def delete_content(self):
+    def init_copy_and_paste_actions(self):
+        """Initializes copy and paste actions and connects relevant signals."""
+        if self._copy_action is not None or self._paste_action is not None:
+            raise RuntimeError("Copy and paste actions have already been set.")
+        copy_icon = QIcon(":/icons/menu_icons/copy.svg")
+        self._copy_action = QAction(copy_icon, "Copy", self)
+        self._copy_action.setShortcut(QKeySequence.Copy)
+        self.addAction(self._copy_action)
+        self._copy_action.triggered.connect(self.copy)
+        paste_icon = QIcon(":/icons/menu_icons/paste.svg")
+        self._paste_action = QAction(paste_icon, "Paste", self)
+        self._paste_action.setShortcut(QKeySequence.Paste)
+        self.addAction(self._paste_action)
+        self._paste_action.triggered.connect(self.paste)
+
+    def set_external_copy_and_paste_actions(self, copy_action, paste_action):
+        """Sets the view to use external copy and paste actions.
+
+        Note that this doesn't connect the actions' trigger signals;
+        the owner of the actions is responsible for handling them.
+
+        Args:
+            copy_action (QAction): copy action
+            paste_action (QAction): paste action
+        """
+        if self._copy_action is not None or self._paste_action is not None:
+            raise RuntimeError("Copy and paste actions have already been set.")
+        self._copy_action = copy_action
+        self._paste_action = paste_action
+
+    @property
+    def copy_action(self):
+        return self._copy_action
+
+    @property
+    def paste_action(self):
+        return self._paste_action
+
+    @Slot(bool)
+    def delete_content(self, _=False):
         """Deletes content from editable indexes in current selection."""
         selection = self.selectionModel().selection()
         if not selection:
@@ -71,7 +105,8 @@ class CopyPasteTableView(QTableView):
         return not self.selectionModel().selection().isEmpty()
 
     @busy_effect
-    def copy(self):
+    @Slot(bool)
+    def copy(self, _=False):
         """Copies current selection to clipboard in excel format."""
         selection = self.selectionModel().selection()
         if not selection:
@@ -111,7 +146,8 @@ class CopyPasteTableView(QTableView):
         )
 
     @busy_effect
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=False):
         """Paste data from clipboard."""
         selection = self.selectionModel().selection()
         if len(selection.indexes()) > 1:
@@ -295,18 +331,26 @@ class AutoFilterCopyPasteTableView(CopyPasteTableView):
 class IndexedParameterValueTableViewBase(CopyPasteTableView):
     """Custom QTableView base class with copy and paste methods for indexed parameter values."""
 
-    def copy(self):
-        """Copies current selection to clipboard in CSV format."""
+    @Slot(bool)
+    def copy(self, _=False):
+        """Copies current selection to clipboard in CSV format.
+
+        Returns:
+            bool: True if data was copied on the clipboard, False otherwise
+        """
         selection_model = self.selectionModel()
         if not selection_model.hasSelection():
             return False
-        selected_indexes = sorted(selection_model.selectedIndexes(), key=lambda index: 2 * index.row() + index.column())
+        data_model = self.model()
+        selected_indexes = sorted(
+            (index for index in selection_model.selectedIndexes() if not data_model.is_expanse_row(index.row())),
+            key=lambda i: 2 * i.row() + i.column(),
+        )
         row_first = selected_indexes[0].row()
         row_last = selected_indexes[-1].row()
         row_count = row_last - row_first + 1
         data_indexes = row_count * [None]
         data_values = row_count * [None]
-        data_model = self.model()
         for selected_index in selected_indexes:
             data = data_model.data(selected_index)
             row = selected_index.row()
@@ -336,7 +380,8 @@ class IndexedParameterValueTableViewBase(CopyPasteTableView):
         """Reads CSV formatted table."""
         raise NotImplementedError()
 
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=False):
         """Pastes data from clipboard to selection."""
         raise NotImplementedError()
 
@@ -344,7 +389,8 @@ class IndexedParameterValueTableViewBase(CopyPasteTableView):
 class TimeSeriesFixedResolutionTableView(IndexedParameterValueTableViewBase):
     """A QTableView for fixed resolution time series table."""
 
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=True):
         """Pastes data from clipboard."""
         selection_model = self.selectionModel()
         if not selection_model.hasSelection():
@@ -420,7 +466,8 @@ class TimeSeriesFixedResolutionTableView(IndexedParameterValueTableViewBase):
 class IndexedValueTableView(IndexedParameterValueTableViewBase):
     """A QTableView class with for variable resolution time series and time patterns."""
 
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=False):
         """Pastes data from clipboard."""
         selection_model = self.selectionModel()
         if not selection_model.hasSelection():
@@ -552,8 +599,13 @@ class IndexedValueTableView(IndexedParameterValueTableViewBase):
 class ArrayTableView(IndexedParameterValueTableViewBase):
     """Custom QTableView with copy and paste methods for single column tables."""
 
-    def copy(self):
-        """Copies current selection to clipboard in CSV format."""
+    @Slot(bool)
+    def copy(self, _=False):
+        """Copies current selection to clipboard in CSV format.
+
+        Returns:
+            bool: True if data was copied on the clipboard, False otherwise
+        """
         selection_model = self.selectionModel()
         if not selection_model.hasSelection():
             return False
@@ -570,7 +622,8 @@ class ArrayTableView(IndexedParameterValueTableViewBase):
                 QApplication.clipboard().setText(output.getvalue())
         return True
 
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=False):
         """Pastes data from clipboard."""
         selection_model = self.selectionModel()
         if not selection_model.hasSelection():
@@ -631,13 +684,25 @@ class ArrayTableView(IndexedParameterValueTableViewBase):
 class MapTableView(CopyPasteTableView):
     """Custom QTableView with copy and paste methods for map tables."""
 
-    def copy(self):
-        """Copies current selection to clipboard in Excel compatible CSV format."""
+    @Slot(bool)
+    def copy(self, _=False):
+        """Copies current selection to clipboard in Excel compatible CSV format.
+
+        Returns:
+            bool: True if data was copied on the clipboard, False otherwise
+        """
         selection = self.selectionModel().selection()
         if not selection:
             return False
         top, bottom, left, right = _range(selection)
         model = self.model()
+        if model.is_expanse_column(right):
+            right -= 1
+        if model.is_expanse_row(bottom):
+            bottom -= 1
+        if left > right or top > bottom:
+            QApplication.clipboard().setText("")
+            return True
         out_table = list()
         with system_lc_numeric():
             for y in range(top, bottom + 1):
@@ -665,14 +730,16 @@ class MapTableView(CopyPasteTableView):
             QApplication.clipboard().setText(output.getvalue())
         return True
 
-    def delete_content(self):
+    @Slot(bool)
+    def delete_content(self, _=False):
         """Deletes content in current selection."""
         selection = self.selectionModel().selection()
         if not selection:
             return False
         self.model().clear(selection.indexes())
 
-    def paste(self):
+    @Slot(bool)
+    def paste(self, _=False):
         """Pastes data from clipboard.
 
         Returns:
@@ -741,7 +808,8 @@ class MapTableView(CopyPasteTableView):
                         except ValueError:
                             pass
                         try:
-                            # Try parsing Duration before DateTime because DateTime will happily accept strings like '1h'
+                            # Try parsing Duration before DateTime
+                            # because DateTime will happily accept strings like '1h'
                             value = Duration(cell)
                             data_row.append(value)
                             continue
