@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -15,10 +15,12 @@ Context menus for parameter value editor widgets.
 :author: A. Soininen (VTT)
 :date:   5.7.2019
 """
+from operator import itemgetter
+
 from PySide2.QtCore import Slot
 from PySide2.QtWidgets import QInputDialog, QMenu, QAction
-from spinetoolbox.plotting import plot_selection, PlottingError, MapTablePlottingHints
-from spinetoolbox.widgets.plot_widget import PlotWidget, _prepare_plot_in_window_menu
+from spinetoolbox.plotting import PlottingError, plot_value_editor_table_selection
+from spinetoolbox.widgets.plot_widget import PlotWidget, prepare_plot_in_window_menu
 from spinetoolbox.widgets.report_plotting_failure import report_plotting_failure
 
 _INSERT_SINGLE_COLUMN_AFTER = "Insert column after"
@@ -53,6 +55,9 @@ class ContextMenuBase(QMenu):
 
     def _add_default_actions(self):
         """Adds default actions to the menu."""
+        self.addAction(self._table_view.copy_action)
+        self.addAction(self._table_view.paste_action)
+        self.addSeparator()
         self.addAction(_INSERT_SINGLE_ROW_BEFORE, self._insert_single_row_before)
         self.addAction(_INSERT_MULTIPLE_ROWS_BEFORE, self._insert_multiple_rows_before)
         self.addSeparator()
@@ -177,7 +182,7 @@ class MapTableContextMenu(ContextMenuBase):
         self.addAction(_PLOT, self._plot)
         self._plot_in_window_menu = self.addMenu(_PLOT_IN_WINDOW)
         self._plot_in_window_menu.triggered.connect(self._plot_in_window)
-        _prepare_plot_in_window_menu(self._plot_in_window_menu)
+        prepare_plot_in_window_menu(self._plot_in_window_menu)
         self.addSeparator()
         self._add_default_actions()
         self.addSeparator()
@@ -263,8 +268,7 @@ class MapTableContextMenu(ContextMenuBase):
         """Plots current indexes."""
         selection = self._table_view.selectedIndexes()
         try:
-            hints = MapTablePlottingHints()
-            plot_widget = plot_selection(self._table_view.model(), selection, hints)
+            plot_widget = plot_value_editor_table_selection(self._table_view.model(), selection)
         except PlottingError as error:
             report_plotting_failure(error, self._table_view)
         else:
@@ -280,9 +284,8 @@ class MapTableContextMenu(ContextMenuBase):
             self._plot()
             return
         selected_indexes = self._table_view.selectedIndexes()
-        hints = MapTablePlottingHints()
         try:
-            plot_selection(self._table_view.model(), selected_indexes, hints, plot_window)
+            plot_value_editor_table_selection(self._table_view.model(), selected_indexes, plot_window)
             plot_window.raise_()
         except PlottingError as error:
             report_plotting_failure(error, self._table_view)
@@ -329,7 +332,9 @@ def _merge_intervals(intervals):
     Returns:
         list of list: merged intervals in the form [first, last]
     """
-    intervals.sort(key=lambda i: i[0])
+    if not intervals:
+        return []
+    intervals.sort(key=itemgetter(0))
     merged_intervals = [intervals.pop(0)]
     while intervals:
         interval = intervals.pop(0)
