@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -15,18 +15,145 @@ from pygments.styles import get_style_by_name
 from pygments.lexers import get_lexer_by_name
 from pygments.util import ClassNotFound
 from pygments.token import Token
-from PySide2.QtCore import Qt, Slot, QTimer, Signal
-from PySide2.QtWidgets import QTextEdit
-from PySide2.QtGui import QFontDatabase, QTextCharFormat, QFont, QTextCursor, QColor, QTextBlockFormat, QTextOption
+from PySide2.QtCore import Qt, Slot, QTimer, Signal, QRect, QSize
+from PySide2.QtWidgets import QPlainTextEdit, QSizePolicy
+from PySide2.QtGui import (
+    QFontDatabase,
+    QTextCharFormat,
+    QFont,
+    QTextCursor,
+    QColor,
+    QTextBlockFormat,
+    QTextOption,
+    QKeySequence,
+)
 from spinetoolbox.helpers import CustomSyntaxHighlighter
 from spinetoolbox.spine_engine_manager import make_engine_manager
+from spinetoolbox.server.engine_client import RemoteEngineInitFailed
+from spinetoolbox.qthread_pool_executor import QtBasedThreadPoolExecutor
 
 
-class PersistentConsoleWidget(QTextEdit):
+class _CustomLineEdit(QPlainTextEdit):
+    def __init__(self, console):
+        super().__init__(console)
+        self._updating = False
+        self._console = console
+        self._current_prompt = ""
+        self.setStyleSheet("QPlainTextEdit {background-color: transparent; color: transparent; border:none;}")
+        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
+        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
+        self.document().setDocumentMargin(0)
+        self.setAttribute(Qt.WA_TransparentForMouseEvents)
+        self.setTabChangesFocus(False)
+        self.setUndoRedoEnabled(False)
+        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
+        self.cursorPositionChanged.connect(self._handle_cursor_position_changed)
+        self.textChanged.connect(self._handle_text_changed)
+
+    @property
+    def min_pos(self):
+        return len(self._current_prompt)
+
+    @property
+    def new_line_indent(self):
+        return len(self._current_prompt.lstrip())  # lstrip() is to remove leading '\n'
+
+    def reset(self, current_prompt):
+        self._current_prompt = current_prompt
+        self.setPlainText(current_prompt)
+
+    def new_line(self):
+        cursor = self.textCursor()
+        cursor.insertText("\n")
+
+    def formatted_text(self):
+        text = self.raw_text()
+        if not text:
+            return ""
+        lines = iter(text.splitlines())
+        new_lines = [next(lines).rstrip()] + [line.rstrip()[self.new_line_indent :] for line in lines]
+        return "\n".join(new_lines)
+
+    def raw_text(self):
+        return self.toPlainText()[self.min_pos :]
+
+    def set_raw_text(self, text):
+        self.setPlainText(self._current_prompt + text)
+
+    @Slot()
+    def _handle_text_changed(self):
+        """Add indent to new lines."""
+        if self._updating:
+            return
+        if not self.raw_text():
+            return
+        self._updating = True
+        cursor = self.textCursor()
+        for i in range(self.document().blockCount()):
+            block = self.document().findBlockByNumber(i)
+            if block.position() < self.min_pos:
+                continue
+            if not block.text().startswith(self.new_line_indent * " "):
+                cursor.setPosition(block.position())
+                cursor.insertText(self.new_line_indent * " ")
+        self._updating = False
+
+    @Slot()
+    def _handle_cursor_position_changed(self):
+        """Move cursor away from indent areas."""
+        if self._updating:
+            return
+        self._updating = True
+        cursor = self.textCursor()
+        if cursor.position() < self.min_pos:
+            cursor.setPosition(self.min_pos)
+        elif cursor.positionInBlock() < self.new_line_indent:
+            cursor.movePosition(QTextCursor.StartOfBlock)
+            cursor.movePosition(QTextCursor.NextCharacter, n=self.new_line_indent)
+        self.setTextCursor(cursor)
+        self._updating = False
+
+    def sizeHint(self):
+        return QSize(self._console.width(), 16777215)
+
+    def keyPressEvent(self, ev):
+        if ev.matches(QKeySequence.Copy):
+            ev.ignore()
+            return
+        if ev.key() == Qt.Key_Backspace:
+            cursor = self.textCursor()
+            if cursor.position() == self.min_pos:
+                return
+            if cursor.positionInBlock() == self.new_line_indent:
+                cursor.movePosition(QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor, n=self.new_line_indent + 1)
+                cursor.removeSelectedText()
+                return
+        if ev.key() == Qt.Key_Left:
+            cursor = self.textCursor()
+            if cursor.positionInBlock() == self.new_line_indent:
+                cursor.movePosition(QTextCursor.PreviousCharacter, n=self.new_line_indent + 1)
+                self.setTextCursor(cursor)
+                return
+        if ev.key() == Qt.Key_Delete:
+            cursor = self.textCursor()
+            if cursor.atBlockEnd():
+                cursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, n=self.new_line_indent + 1)
+                cursor.removeSelectedText()
+                return
+        if self._console.key_press_event(ev):
+            return
+        super().keyPressEvent(ev)
+
+
+class PersistentConsoleWidget(QPlainTextEdit):
     """A widget to interact with a persistent process."""
 
+    _command_checked = Signal(str, bool)
+    _msg_available = Signal(str, str)
+    _command_finished = Signal()
     _history_item_available = Signal(str, str)
     _completions_available = Signal(str, str, list)
+    _restarted = Signal()
     _flush_needed = Signal()
     _FLUSH_INTERVAL = 200
     _MAX_LINES_PER_SECOND = 2000
@@ -42,21 +169,17 @@ class PersistentConsoleWidget(QTextEdit):
             owner (ProjectItemBase, optional): console owner
         """
         super().__init__(parent=toolbox)
+        self._executor = QtBasedThreadPoolExecutor(max_workers=1)
+        self._updating = False
         font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
         self.setFont(font)
-        cursor_width = self.fontMetrics().horizontalAdvance("x")
-        self.setCursorWidth(cursor_width)
-        self.document().setIndentWidth(cursor_width)
-        self.document().setMaximumBlockCount(self._MAX_LINES_COUNT)
-        self.setWordWrapMode(QTextOption.WrapAnywhere)
-        self.setTabStopDistance(4 * cursor_width)
+        self.setMaximumBlockCount(self._MAX_LINES_COUNT)
         self._toolbox = toolbox
         self._key = key
         self._language = language
         self.owners = {owner}
         self._prompt, self._prompt_format = self._make_prompt()
         self._prefix = None
-        self._reset_prefix = True
         self._pending_command_count = 0
         self._text_buffer = []
         self._skipped = {}
@@ -65,8 +188,16 @@ class PersistentConsoleWidget(QTextEdit):
         background_color = self._style.background_color
         foreground_color = self._style.styles[Token] or self._style.styles[Token.Text]
         self.setStyleSheet(
-            f"QTextEdit {{background-color: {background_color}; color: {foreground_color}; border: 0px}}"
+            f"QPlainTextEdit {{background-color: {background_color}; color: {foreground_color}; border: 0px}}"
         )
+        cursor_width = self.fontMetrics().horizontalAdvance("x")
+        self.setWordWrapMode(QTextOption.WrapAnywhere)
+        self.setTabStopDistance(4 * cursor_width)
+        self._line_edit = _CustomLineEdit(self)
+        self._line_edit.setFont(font)
+        self._line_edit.setCursorWidth(cursor_width)
+        self._line_edit.setWordWrapMode(QTextOption.WrapAnywhere)
+        self._line_edit.setTabStopDistance(4 * cursor_width)
         self._highlighter = CustomSyntaxHighlighter(None)
         self._highlighter.set_style(self._style)
         try:
@@ -78,25 +209,49 @@ class PersistentConsoleWidget(QTextEdit):
         self._current_prompt = ""
         self._make_prompt_block(prompt=self._prompt)
         self._at_bottom = True
-        self.cursorPositionChanged.connect(self._handle_cursor_position_changed)
-        self.document().contentsChanged.connect(self._handle_contents_changed)
-        self._history_item_available.connect(self._display_history_item)
-        self._completions_available.connect(self._display_completions)
-        self._flush_needed.connect(self._start_flush_timer)
-        self._flush_in_progess = False
+        self._flush_in_progress = False
         self._flush_timer = QTimer()
         self._flush_timer.setInterval(self._FLUSH_INTERVAL)
         self._flush_timer.timeout.connect(self._flush_text_buffer)
         self._flush_timer.setSingleShot(True)
+        self.engine_mngr = None
+        self.setReadOnly(True)
+        self.document().contentsChanged.connect(self._handle_contents_changed)
+        self.updateRequest.connect(self._handle_update_request)
+        self.selectionChanged.connect(self._handle_selection_changed)
+        self.cursorPositionChanged.connect(self._handle_cursor_position_changed)
+        self._flush_needed.connect(self._start_flush_timer)
+        self._line_edit.textChanged.connect(self._update_user_input)
+        self._command_checked.connect(self._handle_command_checked)
+        self._msg_available.connect(self._handle_msg_available)
+        self._command_finished.connect(self._handle_command_finished)
+        self._history_item_available.connect(self._display_history_item)
+        self._completions_available.connect(self._display_completions)
+        self._restarted.connect(self._handle_restarted)
+
+    def closeEvent(self, ev):
+        super().closeEvent(ev)
+        self._executor.shutdown()
 
     def name(self):
         """Returns console name for display purposes."""
         return f"{self._language.capitalize()} Console - {self.owner_names}"
 
+    @property
+    def prompt(self):
+        return self._prompt
+
     @property
     def owner_names(self):
         return " & ".join(x.name for x in self.owners if x is not None)
 
+    @property
+    def _input_start_pos(self):
+        return self._prompt_block.position() + len(self._current_prompt)
+
+    def focusInEvent(self, ev):
+        self._line_edit.setFocus()
+
     def mouseMoveEvent(self, ev):
         super().mouseMoveEvent(ev)
         if self.anchorAt(ev.pos()):
@@ -136,90 +291,58 @@ class PersistentConsoleWidget(QTextEdit):
             scrollbar = self.verticalScrollBar()
             scrollbar.setValue(scrollbar.maximum())
 
-    @property
-    def _input_start_pos(self):
-        return self._prompt_block.position() + len(self._current_prompt)
+    @Slot()
+    def _handle_selection_changed(self):
+        if self._updating:
+            return
+        cursor = self.textCursor()
+        le_cursor = self._line_edit.textCursor()
+        le_selection_start = cursor.selectionStart() - self._prompt_block.position()
+        le_selection_end = cursor.selectionEnd() - self._prompt_block.position()
+        if le_selection_start < self._line_edit.min_pos and le_selection_end < self._line_edit.min_pos:
+            le_cursor.clearSelection()
+        else:
+            le_selection_start = max(self._line_edit.min_pos, le_selection_start)
+            le_selection_end = max(self._line_edit.min_pos, le_selection_end)
+            le_cursor.setPosition(le_selection_start)
+            le_cursor.setPosition(le_selection_end, QTextCursor.KeepAnchor)
+        self._line_edit.setTextCursor(le_cursor)
 
     @Slot()
     def _handle_cursor_position_changed(self):
-        self.setReadOnly(self.textCursor().position() < self._input_start_pos)
-        if self._reset_prefix:
-            self._prefix = None
-
-    def _make_prompt(self):
-        text_format = QTextCharFormat()
-        if self._language == "julia":
-            prompt = "\njulia> "
-            text_format.setForeground(Qt.darkGreen)
-            text_format.setFontWeight(QFont.Bold)
-        elif self._language == "python":
-            prompt = ">>> "
-        else:
-            prompt = "$ "
-        return prompt, text_format
-
-    def _make_prompt_block(self, prompt=""):
+        if self._updating:
+            return
         cursor = self.textCursor()
-        cursor.movePosition(QTextCursor.End)
-        cursor.insertBlock(QTextBlockFormat())
-        self._prompt_block = cursor.block()
-        self._insert_prompt(prompt=prompt)
+        le_cursor = self._line_edit.textCursor()
+        le_position = cursor.position() - self._prompt_block.position()
+        if self._line_edit.min_pos <= le_position < self._line_edit.document().characterCount():
+            le_cursor.setPosition(le_position)
+            self._line_edit.setTextCursor(le_cursor)
 
-    def _insert_prompt(self, prompt=""):
+    @Slot(QRect, int)
+    def _handle_update_request(self, _rect, _dy):
+        """Move line edit to input start pos."""
+        if not self._updating:
+            self._move_line_edit()
+
+    def _move_line_edit(self):
+        if self._prompt_block is None:
+            return
         cursor = self.textCursor()
         cursor.setPosition(self._prompt_block.position())
-        cursor.insertText(prompt, self._prompt_format)
-        cursor.movePosition(QTextCursor.End)
-        self._current_prompt = prompt
-        self.setTextCursor(cursor)
+        rect = self.cursorRect(cursor)
+        self._line_edit.move(rect.topLeft())
 
-    def _make_continuation_block(self, cursor):
-        block_format = QTextBlockFormat()
-        block_format.setIndent(len(self._current_prompt.lstrip()))
-        cursor.insertBlock(block_format)
-
-    def _insert_stdin_text(self, cursor, text):
-        """Inserts highlighted text.
-
-        Args:
-            cursor (QTextCursor)
-            text (str)
-        """
-        if not text:
-            cursor.insertText("")
-            return
-        lines = iter(text.splitlines())
-        line = next(lines)
-        self._do_insert_stdin_text(cursor, line)
-        for line in lines:
-            self._make_continuation_block(cursor)
-            self._do_insert_stdin_text(cursor, line)
-
-    def _do_insert_stdin_text(self, cursor, text):
-        for start, count, text_format in self._highlighter.yield_formats(text):
-            chunk = text[start : start + count]
-            cursor.insertText(chunk, text_format)
-
-    def _insert_stdout_text(self, cursor, text):
-        """Inserts ansi highlighted text.
-
-        Args:
-            cursor (QTextCursor)
-            text (str)
-        """
-        for chunk, text_format in self._ansi_esc_code_handler.parse_text(text):
-            cursor.insertText(chunk, text_format)
-
-    def _insert_text_before_prompt(self, text, with_prompt=False):
-        """Inserts given text before the prompt. Used when adding input and output from external execution.
-
-        Args:
-            text (str)
-        """
-        self._text_buffer.append((text, with_prompt))
-        if not self._flush_in_progess:
-            self._flush_in_progess = True
-            self._flush_needed.emit()
+    @Slot()
+    def _update_user_input(self):
+        self._updating = True
+        cursor = self.textCursor()
+        cursor.setPosition(self._input_start_pos)
+        cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
+        text = self._line_edit.raw_text()
+        cursor.insertText(text)
+        self._highlight_current_input()
+        self._updating = False
 
     @Slot()
     def _start_flush_timer(self):
@@ -249,7 +372,78 @@ class PersistentConsoleWidget(QTextEdit):
             cursor.insertText(f"<--- {len(self._text_buffer)} more lines --->", char_format)
             self._text_buffer.clear()
         cursor.endEditBlock()
-        self._flush_in_progess = False
+        self._flush_in_progress = False
+
+    def _make_prompt(self):
+        text_format = QTextCharFormat()
+        if self._language == "julia":
+            prompt = "\njulia> "
+            text_format.setForeground(Qt.darkGreen)
+            text_format.setFontWeight(QFont.Bold)
+        elif self._language == "python":
+            prompt = ">>> "
+        else:
+            prompt = "$ "
+        return prompt, text_format
+
+    def _make_prompt_block(self, prompt=""):
+        cursor = self.textCursor()
+        cursor.movePosition(QTextCursor.End)
+        cursor.insertBlock()
+        self._prompt_block = cursor.block()
+        self._insert_prompt(prompt=prompt)
+
+    def _insert_prompt(self, prompt=""):
+        cursor = self.textCursor()
+        cursor.setPosition(self._prompt_block.position())
+        cursor.insertText(prompt, self._prompt_format)
+        cursor.movePosition(QTextCursor.End)
+        self._current_prompt = prompt
+        self._line_edit.reset(self._current_prompt)
+        self._move_line_edit()
+
+    def _insert_stdin_text(self, cursor, text):
+        """Inserts highlighted text.
+
+        Args:
+            cursor (QTextCursor)
+            text (str)
+        """
+        if not text:
+            cursor.insertText("")
+            return
+        lines = iter(text.splitlines())
+        line = next(lines)
+        self._do_insert_stdin_text(cursor, line)
+        for line in lines:
+            cursor.insertText("\n")
+            self._do_insert_stdin_text(cursor, line)
+
+    def _do_insert_stdin_text(self, cursor, text):
+        for start, count, text_format in self._highlighter.yield_formats(text):
+            chunk = text[start : start + count]
+            cursor.insertText(chunk, text_format)
+
+    def _insert_stdout_text(self, cursor, text):
+        """Inserts ansi highlighted text.
+
+        Args:
+            cursor (QTextCursor)
+            text (str)
+        """
+        for chunk, text_format in self._ansi_esc_code_handler.parse_text(text):
+            cursor.insertText(chunk, text_format)
+
+    def _insert_text_before_prompt(self, text, with_prompt=False):
+        """Inserts given text before the prompt. Used when adding input and output from external execution.
+
+        Args:
+            text (str)
+        """
+        self._text_buffer.append((text, with_prompt))
+        if not self._flush_in_progress:
+            self._flush_in_progress = True
+            self._flush_needed.emit()
 
     def _insert_text(self, cursor, text, with_prompt):
         cursor.insertBlock(QTextBlockFormat())
@@ -285,22 +479,12 @@ class PersistentConsoleWidget(QTextEdit):
         self._insert_text_before_prompt(data)
 
     def _get_current_text(self):
-        """Returns current text.
-
-        Returns:
-            str: the complete text
-        """
-        cursor = self.textCursor()
-        cursor.setPosition(self._input_start_pos)
-        cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
-        return cursor.selectedText()
+        return self._line_edit.formatted_text()
 
     def _get_prefix(self):
-        if self._prefix is None:
-            cursor = self.textCursor()
-            cursor.setPosition(self._input_start_pos, QTextCursor.KeepAnchor)
-            self._prefix = cursor.selectedText()
-        return self._prefix
+        le_cursor = self._line_edit.textCursor()
+        le_cursor.setPosition(self._line_edit.min_pos, QTextCursor.KeepAnchor)
+        return le_cursor.selectedText().rstrip()
 
     def _highlight_current_input(self):
         cursor = self.textCursor()
@@ -315,7 +499,12 @@ class PersistentConsoleWidget(QTextEdit):
         cursor.movePosition(QTextCursor.NextCharacter)
         cursor.setCharFormat(QTextCharFormat())
 
-    def keyPressEvent(self, ev):
+    def key_press_event(self, ev):
+        """Handles key press event from line edit.
+
+        Returns:
+            True if handled, False if not.
+        """
         self._at_bottom = True
         text = self._get_current_text()
         if ev.key() in (Qt.Key_Return, Qt.Key_Enter):
@@ -326,13 +515,29 @@ class PersistentConsoleWidget(QTextEdit):
             self._move_history(text, False)
         elif ev.key() == Qt.Key_Tab:
             self._autocomplete(text)
-        elif ev.key() != Qt.Key_Backspace or self.textCursor().position() > self._input_start_pos:
-            cursor = self.textCursor()
-            super().keyPressEvent(ev)
-            if cursor.position() >= self._input_start_pos and self.textCursor().position() < self._input_start_pos:
-                cursor.setPosition(self._input_start_pos)
-                self.setTextCursor(cursor)
-            self._highlight_current_input()
+        else:
+            return False
+        return True
+
+    def create_engine_manager(self):
+        """Returns a new local or remote spine engine manager or
+        an existing remote spine engine manager.
+        Returns None if connecting to Spine Engine Server fails."""
+        exec_remotely = self._toolbox.qsettings().value("engineSettings/remoteExecutionEnabled", "false") == "true"
+        if exec_remotely:
+            if self.engine_mngr:
+                return self.engine_mngr
+            self.engine_mngr = make_engine_manager(exec_remotely)
+            host, port, security, sec_folder = self._toolbox.engine_server_settings()
+            try:
+                self.engine_mngr.make_engine_client(host, port, security, sec_folder)
+            except RemoteEngineInitFailed as e:
+                self._toolbox.msg_error.emit(f"Connecting to Spine Engine Server failed. {e}")
+                return None
+            return self.engine_mngr
+        else:
+            engine_mngr = make_engine_manager(exec_remotely)
+            return engine_mngr
 
     def _issue_command(self, text):
         """Issues command.
@@ -340,27 +545,52 @@ class PersistentConsoleWidget(QTextEdit):
         Args:
             text (str)
         """
-        engine_server_address = self._toolbox.qsettings().value("appSettings/engineServerAddress", defaultValue="")
-        engine_mngr = make_engine_manager(engine_server_address)
-        if not engine_mngr.is_persistent_command_complete(self._key, text):
-            self._make_continuation_block(self.textCursor())
+        self._executor.submit(self._do_check_command, text)
+
+    def _do_check_command(self, text):
+        if not text.strip():  # Don't send empty command to execution manager
+            self._make_prompt_block(prompt=self._prompt)
+            return
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
+            return
+        complete = engine_mngr.is_persistent_command_complete(self._key, text)
+        self._command_checked.emit(text, complete)
+
+    @Slot(str, bool)
+    def _handle_command_checked(self, text, complete):
+        """Issues command.
+
+        Args:
+            text (str)
+        """
+        if not complete:
+            self._line_edit.new_line()
+            return
+        self._make_prompt_block(prompt="")
+        self._executor.submit(self._do_issue_command, text)
+
+    def _do_issue_command(self, text):
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
             return
         log_stdin = bool(self._pending_command_count)
-        self._make_prompt_block(prompt="")
         self._pending_command_count += 1
-        self._do_issue_command(engine_mngr, text, log_stdin)
-
-    def _do_issue_command(self, engine_mngr, text, log_stdin):
         for msg in engine_mngr.issue_persistent_command(self._key, text):
-            msg_type = msg["type"]
-            if msg_type == "stdin" and log_stdin:
-                self.add_stdin(msg["data"])
-            elif msg_type == "stdout":
-                self.add_stdout(msg["data"])
-            elif msg_type == "stderr":
-                self.add_stderr(msg["data"])
-        self._handle_command_finished()
+            if msg["type"] != "stdin" or log_stdin:
+                self._msg_available.emit(msg["type"], msg["data"])
+        self._command_finished.emit()
 
+    @Slot(str, str)
+    def _handle_msg_available(self, msg_type, text):
+        if msg_type == "stdin":
+            self.add_stdin(text)
+        elif msg_type == "stdout":
+            self.add_stdout(text)
+        elif msg_type == "stderr":
+            self.add_stderr(text)
+
+    @Slot()
     def _handle_command_finished(self):
         self._pending_command_count -= 1
         if self._pending_command_count == 0:
@@ -368,25 +598,23 @@ class PersistentConsoleWidget(QTextEdit):
 
     def _move_history(self, text, backwards):
         """Moves history."""
-        engine_server_address = self._toolbox.qsettings().value("appSettings/engineServerAddress", defaultValue="")
-        engine_mngr = make_engine_manager(engine_server_address)
+        self._executor.submit(self._do_move_history, text, backwards)
+
+    def _do_move_history(self, text, backwards):
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
+            return
         prefix = self._get_prefix()
         history_item = engine_mngr.get_persistent_history_item(self._key, text, prefix, backwards)
         self._history_item_available.emit(history_item, prefix)
 
     @Slot(str, str)
     def _display_history_item(self, history_item, prefix):
-        self._reset_prefix = False
-        cursor = self.textCursor()
-        pos = self._input_start_pos + len(prefix)
-        cursor.setPosition(pos)
-        cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
-        self._insert_stdin_text(cursor, history_item[len(prefix) :])
+        self._line_edit.set_raw_text(history_item)
         if prefix:
-            cursor.setPosition(pos)
-        self.setTextCursor(cursor)
-        self._reset_prefix = True
-        self._highlight_current_input()
+            le_cursor = self._line_edit.textCursor()
+            le_cursor.setPosition(self._line_edit.min_pos + len(prefix))
+            self._line_edit.setTextCursor(le_cursor)
 
     def _autocomplete(self, text):
         """Autocompletes current text in the prompt (or output options if multiple matches).
@@ -394,8 +622,18 @@ class PersistentConsoleWidget(QTextEdit):
         Args:
             text (str)
         """
-        engine_server_address = self._toolbox.qsettings().value("appSettings/engineServerAddress", defaultValue="")
-        engine_mngr = make_engine_manager(engine_server_address)
+        prev_char = self._line_edit.document().characterAt(self._line_edit.textCursor().position() - 1)
+        if prev_char.isspace():
+            le_cursor = self._line_edit.textCursor()
+            le_cursor.insertText(4 * " ")
+            self._line_edit.setTextCursor(le_cursor)
+            return
+        self._executor.submit(self._do_autocomplete, text)
+
+    def _do_autocomplete(self, text):
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
+            return
         prefix = self._get_prefix()
         completions = engine_mngr.get_persistent_completions(self._key, prefix)
         self._completions_available.emit(text, prefix, completions)
@@ -407,33 +645,51 @@ class PersistentConsoleWidget(QTextEdit):
             # Can't complete, but there is more than one option: 'commit' stdin and output options to stdout
             self.add_stdin(text)
             self.add_stdout("\t\t".join(completions))
+            le_cursor = self._line_edit.textCursor()
+            self._line_edit.set_raw_text(text)
+            self._line_edit.setTextCursor(le_cursor)
         else:
             # Complete in current line
-            cursor = self.textCursor()
             last_prefix_word = prefix.split(" ")[-1]
-            cursor.insertText(completion[len(last_prefix_word) :])
-            self._highlight_current_input()
+            text_to_insert = completion[len(last_prefix_word) :]
+            index = len(prefix)
+            new_text = text[:index] + text_to_insert + text[index:]
+            self._line_edit.set_raw_text(new_text)
+            le_cursor = self._line_edit.textCursor()
+            le_cursor.setPosition(self._line_edit.min_pos + index + len(text_to_insert))
+            self._line_edit.setTextCursor(le_cursor)
 
     @Slot(bool)
     def _restart_persistent(self, _=False):
         """Restarts underlying persistent process."""
+        self._updating = True
         self.clear()
-        engine_server_address = self._toolbox.qsettings().value("appSettings/engineServerAddress", defaultValue="")
-        engine_mngr = make_engine_manager(engine_server_address)
+        self._make_prompt_block("")
+        self._updating = False
         self._text_buffer.clear()
+        self._executor.submit(self._do_restart_persistent)
+
+    def _do_restart_persistent(self):
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
+            return
         for msg in engine_mngr.restart_persistent(self._key):
-            msg_type = msg["type"]
-            if msg_type == "stdout":
-                self.add_stdout(msg["data"])
-            elif msg_type == "stderr":
-                self.add_stderr(msg["data"])
+            self._msg_available.emit(msg["type"], msg["data"])
+        self._restarted.emit()
+
+    @Slot()
+    def _handle_restarted(self):
         self._make_prompt_block(prompt=self._prompt)
 
     @Slot(bool)
     def _interrupt_persistent(self, _=False):
         """Interrupts underlying persistent process."""
-        engine_server_address = self._toolbox.qsettings().value("appSettings/engineServerAddress", defaultValue="")
-        engine_mngr = make_engine_manager(engine_server_address)
+        self._executor.submit(self._do_interrupt_persistent)
+
+    def _do_interrupt_persistent(self):
+        engine_mngr = self.create_engine_manager()
+        if not engine_mngr:
+            return
         engine_mngr.interrupt_persistent(self._key)
 
     def _extend_menu(self, menu):
@@ -442,11 +698,16 @@ class PersistentConsoleWidget(QTextEdit):
         menu.addAction("Restart", self._restart_persistent)
         menu.addAction("Interrupt", self._interrupt_persistent)
 
-    def contextMenuEvent(self, event):
+    def contextMenuEvent(self, ev):
         """Reimplemented to extend menu with custom actions."""
-        menu = self.createStandardContextMenu()
+        le_geom = self._line_edit.frameGeometry()
+        menu = (
+            self._line_edit.createStandardContextMenu()
+            if le_geom.contains(ev.pos())
+            else self.createStandardContextMenu()
+        )
         self._extend_menu(menu)
-        menu.exec_(event.globalPos())
+        menu.exec_(ev.globalPos())
 
 
 # Translated from
