@@ -1,5 +1,5 @@
 ######################################################################################################################
-# Copyright (C) 2017-2021 Spine project consortium
+# Copyright (C) 2017-2022 Spine project consortium
 # This file is part of Spine Toolbox.
 # Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 # Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
@@ -16,7 +16,7 @@ Classes for custom QListView.
 :date:   14.11.2018
 """
 
-import textwrap
+from textwrap import fill
 from PySide2.QtCore import QModelIndex, Qt, Signal, Slot, QMimeData
 from PySide2.QtGui import QDrag, QIcon, QPainter, QBrush, QColor, QFont, QIconEngine
 from PySide2.QtWidgets import QToolButton, QApplication, QToolBar, QWidgetAction
@@ -61,7 +61,25 @@ class ProjectItemDragMixin:
         self.mime_data = None
 
 
-class ProjectItemButtonBase(ProjectItemDragMixin, QToolButton):
+class NiceButton(QToolButton):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        font = self.font()
+        font.setPointSize(9)
+        self.setFont(font)
+        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
+
+    def setText(self, text):
+        super().setText(fill(text, width=12, break_long_words=False))
+
+    def set_orientation(self, orientation):
+        if orientation == Qt.Horizontal:
+            self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
+        else:
+            self.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+
+
+class ProjectItemButtonBase(ProjectItemDragMixin, NiceButton):
     def __init__(self, toolbox, item_type, icon, parent=None):
         super().__init__(parent=parent)
         self._toolbox = toolbox
@@ -99,6 +117,7 @@ class ProjectItemButton(ProjectItemButtonBase):
     def __init__(self, toolbox, item_type, icon, parent=None):
         super().__init__(toolbox, item_type, icon, parent=parent)
         self.setToolTip(f"<p>Drag-and-drop this onto the Design View to create a new <b>{item_type}</b> item.</p>")
+        self.setText(item_type)
 
     def _make_mime_data_text(self):
         return ",".join([self.item_type, ""])
@@ -113,18 +132,7 @@ class ProjectItemSpecButton(ProjectItemButtonBase):
         self._spec_name = None
         self._index = None
         self.spec_name = spec_name
-        font = self.font()
-        font.setPointSize(9)
-        self.setFont(font)
         self.setText(self.spec_name)
-        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
-
-    def set_orientation(self, orientation):
-        if orientation == Qt.Horizontal:
-            new_text = self.spec_name
-        else:
-            new_text = textwrap.fill(self.spec_name, width=8)
-        self.setText(new_text)
 
     @property
     def spec_name(self):
@@ -163,7 +171,7 @@ class ShadeProjectItemSpecButton(ShadeMixin, ProjectItemSpecButton):
         return ShadeProjectItemSpecButton(self._toolbox, self.item_type, self.icon(), self.spec_name)
 
 
-class ShadeButton(ShadeMixin, QToolButton):
+class ShadeButton(ShadeMixin, NiceButton):
     pass
 
 
@@ -223,12 +231,9 @@ class ProjectItemSpecArray(QToolBar):
         self.addWidget(self._button_visible)
         self._button_new = ShadeButton()
         self._button_new.setIcon(QIcon(CharIconEngine("\uf067", color=self._icon.color())))
+        self._button_new.setIconSize(self.iconSize())
         self._button_new.setText("New...")
         self._button_new.setToolTip(f"<p>Create new <b>{item_type}</b> specification...</p>")
-        font = QFont()
-        font.setPointSize(9)
-        self._button_new.setFont(font)
-        self._button_new.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
         self._action_new = self.addWidget(self._button_new)
         self._action_new.setVisible(self._visible)
         self._actions = {}
@@ -276,7 +281,8 @@ class ProjectItemSpecArray(QToolBar):
             list(QAction)
             int or NoneType
         """
-        actions = [*self._actions.values()]
+        actions_iter = (self._actions.get(spec.name) for spec in self._model.specifications())
+        actions = [act for act in actions_iter if act is not None]
         if self.orientation() == Qt.Horizontal:
             get_point = lambda ref_geom: (ref_geom.right() + 1, ref_geom.top())
         else:
@@ -378,6 +384,8 @@ class ProjectItemSpecArray(QToolBar):
         spacing = 2  # additional space till next toolbar icon when collapsed
         if orientation is None:
             orientation = self.orientation()
+        self._button_base_item.set_orientation(orientation)
+        self._button_new.set_orientation(orientation)
         widgets = [self.widgetForAction(a) for a in self._actions.values()]
         for w in widgets:
             w.set_orientation(orientation)
@@ -387,11 +395,11 @@ class ProjectItemSpecArray(QToolBar):
         if orientation == Qt.Horizontal:
             icon = down if not self._visible else right
             width = extent
-            min_width = self._button_visible.frameGeometry().right() + self._margin + spacing
+            min_width = self._button_base_item.sizeHint().width() + extent + self._margin + spacing
             min_visible_width = min_width + self._button_new.sizeHint().width() - spacing
             if widgets:
                 min_visible_width += extent
-            min_height = self._button_new.sizeHint().height()
+            min_height = self._button_base_item.sizeHint().height()
             min_size = (min_width, min_height)
             min_visible_size = (min_visible_width, min_height)
             height = max((w.sizeHint().height() for w in widgets), default=min_height)
@@ -402,8 +410,8 @@ class ProjectItemSpecArray(QToolBar):
         else:
             icon = right if not self._visible else down
             height = extent
-            min_width = self._button_new.sizeHint().height()
-            min_height = self._button_visible.frameGeometry().bottom() + self._margin + spacing
+            min_width = self._button_base_item.sizeHint().width()
+            min_height = self._button_base_item.sizeHint().height() + extent + self._margin + spacing
             min_visible_height = min_height + self._button_new.sizeHint().height() - spacing
             if widgets:
                 min_visible_height += extent
